"""OpenAI engine interface."""

import re

from typing_extensions import override

from ubo_app.constants.assistant import (
    OPENAI_API_KEY_PATTERN,
    OPENAI_API_KEY_SECRET_ID,
)
from ubo_app.engines.abstraction.ai_provider_mixin import AIProviderMixin
from ubo_app.engines.abstraction.needs_setup_mixin import NeedsSetupMixin
from ubo_app.engines.abstraction.remote_mixin import RemoteMixin
from ubo_app.store.input.types import (
    InputFieldDescription,
    InputFieldType,
    QRCodeInputDescription,
    WebUIInputDescription,
)
from ubo_app.store.services.speech_synthesis import ReadableInformation
from ubo_app.utils import secrets
from ubo_app.utils.input import ubo_input


class OpenAIEngine(NeedsSetupMixin, AIProviderMixin, RemoteMixin):
    """OpenAI engine."""

    @property
    def name(self) -> str:
        """The internal name of the OpenAI engine."""
        return 'openai'

    @property
    def label(self) -> str:
        """The display label for the OpenAI engine."""
        return 'OpenAI'

    @property
    def not_setup_message(self) -> str:
        """Message shown when the OpenAI service API key is not set."""
        return 'OpenAI service API key is not set. You can set it in the settings.'

    @property
    @override
    def is_setup(self) -> bool:
        """Check if the OpenAI engine is set up."""
        service_account_info_string = secrets.read_secret(
            OPENAI_API_KEY_SECRET_ID,
        )
        return (
            bool(service_account_info_string)
            and re.match(
                OPENAI_API_KEY_PATTERN,
                service_account_info_string,
            )
            is not None
        )

    async def _setup(self) -> None:
        _, result = await ubo_input(
            title='OpenAI API Key',
            prompt='Enter your OpenAI API key.',
            descriptions=[
                WebUIInputDescription(
                    fields=[
                        InputFieldDescription(
                            name='api_key',
                            type=InputFieldType.TEXT,
                            label='API key',
                            description='Enter your OpenAI API key',
                            required=True,
                            pattern=OPENAI_API_KEY_PATTERN,
                        ),
                    ],
                ),
                QRCodeInputDescription(
                    title='OpenAI API Key',
                    instructions=ReadableInformation(
                        text='Convert your OpenAI API key to a QR code and hold it in '
                        'front of the camera to scan it.',
                        picovoice_text='Convert your Open{AI|EY AY} API key to a '
                        '{QR|K Y UW AA R} code and hold it in front of the camera to '
                        'scan it.',
                    ),
                    pattern=r'(?P<api_key>' + OPENAI_API_KEY_PATTERN + ')',
                ),
            ],
        )
        secrets.write_secret(
            key=OPENAI_API_KEY_SECRET_ID,
            value=result.data['api_key'],
        )



"""Docker store types."""

from __future__ import annotations

import functools
from dataclasses import field
from enum import StrEnum, auto

from immutable import Immutable
from redux import BaseAction, BaseCombineReducerState, BaseEvent

from ubo_app.utils.persistent_store import read_from_persistent_store


class DockerStatus(StrEnum):
    """Docker status."""

    UNKNOWN = auto()
    NOT_INSTALLED = auto()
    INSTALLING = auto()
    NOT_RUNNING = auto()
    RUNNING = auto()
    ERROR = auto()


class DockerItemStatus(StrEnum):
    """Image status."""

    NOT_AVAILABLE = auto()
    FETCHING = auto()
    AVAILABLE = auto()
    CREATED = auto()
    RUNNING = auto()
    ERROR = auto()
    PROCESSING = auto()


class DockerAction(BaseAction):
    """Docker action."""


class DockerInstallAction(DockerAction):
    """Install docker."""


class DockerStartAction(DockerAction):
    """Start docker service."""


class DockerStopAction(DockerAction):
    """Stop docker service."""


class DockerSetStatusAction(DockerAction):
    """Set status of docker service."""

    status: DockerStatus


class DockerStoreUsernameAction(DockerAction):
    """Store username for a registry."""

    registry: str
    username: str


class DockerRemoveUsernameAction(DockerAction):
    """Remove a registry for stored usernames."""

    registry: str


class DockerImageAction(DockerAction):
    """Docker image action."""

    image: str


class DockerImageSetStatusAction(DockerImageAction):
    """Docker image set status action."""

    status: DockerItemStatus
    ports: list[str] | None = None
    ip: str | None = None


class DockerImageSetDockerIdAction(DockerImageAction):
    """Docker image set docker id action."""

    docker_id: str


class DockerImageUpdateMetadataAction(DockerImageAction):
    """Update image metadata (e.g., instructions)."""

    instructions: str | None = None


class DockerImageFetchAction(DockerImageAction):
    """Fetch image or composition."""


class DockerImageRemoveAction(DockerImageAction):
    """Remove image or composition."""


class DockerImageRunAction(DockerImageAction):
    """Run container or composition."""


class DockerImageStopAction(DockerImageAction):
    """Stop container or composition."""


class DockerImageReleaseAction(DockerImageAction):
    """Release composition resources (stop + cleanup)."""


class DockerImageRemoveContainerAction(DockerImageAction):
    """Remove container."""


class DockerEvent(BaseEvent):
    """Docker event."""


class DockerInstallEvent(DockerEvent):
    """Signal for installing docker."""


class DockerStartEvent(DockerEvent):
    """Signal for starting docker service."""


class DockerStopEvent(DockerEvent):
    """Signal for stopping docker service."""


class DockerServiceState(Immutable):
    """Docker service state."""

    status: DockerStatus = DockerStatus.UNKNOWN
    usernames: dict[str, str] = field(
        default_factory=functools.partial(
            read_from_persistent_store,
            'docker_usernames',
            output_type=dict[str, str],
        ),
    )


class DockerImageEvent(DockerEvent):
    """Docker image event."""

    image: str


class DockerImageRegisterAppEvent(DockerImageEvent):
    """Register image entry in apps in store."""


class DockerImageFetchCompositionEvent(DockerImageEvent):
    """Fetch composition."""


class DockerImageFetchEvent(DockerImageEvent):
    """Fetch image."""


class DockerImageRemoveCompositionEvent(DockerImageEvent):
    """Remove composition."""


class DockerImageRemoveEvent(DockerImageEvent):
    """Remove image."""


class DockerImageRunCompositionEvent(DockerImageEvent):
    """Run composition."""


class DockerImageRunContainerEvent(DockerImageEvent):
    """Run container."""


class DockerImageStopCompositionEvent(DockerImageEvent):
    """Stop composition."""


class DockerImageStopContainerEvent(DockerImageEvent):
    """Stop container."""


class DockerImageReleaseCompositionEvent(DockerImageEvent):
    """Release composition."""


class DockerImageRemoveContainerEvent(DockerImageEvent):
    """Remove container."""

class ImageState(Immutable):
    """Image state."""

    id: str
    label: str
    instructions: str | None
    status: DockerItemStatus = DockerItemStatus.NOT_AVAILABLE
    container_ip: str | None = None
    docker_id: str | None = None
    ports: list[str] = field(default_factory=list)

    @property
    def is_fetching(self: ImageState) -> bool:
        """Check if image is available."""
        return self.status == DockerItemStatus.FETCHING

    @property
    def is_available(self: ImageState) -> bool:
        """Check if image is available."""
        return self.status in [DockerItemStatus.AVAILABLE, DockerItemStatus.RUNNING]

    @property
    def is_running(self: ImageState) -> bool:
        """Check if image is running."""
        return self.status == DockerItemStatus.RUNNING


class DockerState(BaseCombineReducerState):
    """Docker state."""

    service: DockerServiceState

    def __getattribute__(self: DockerState, name: str) -> ImageState:
        """Set type for random attributes of DockerState."""
        return super().__getattribute__(name)



"""Constants for the Vosk speech recognition model."""

from ubo_app.constants import DATA_PATH
from ubo_app.store.services.speech_recognition import SpeechRecognitionEngineName

VOSK_MODEL = 'vosk-model-small-en-us-0.15'

VOSK_MODEL_URL = f'https://alphacephei.com/vosk/models/{VOSK_MODEL}.zip'

VOSK_MODEL_PATH = DATA_PATH / VOSK_MODEL
VOSK_DOWNLOAD_PATH = VOSK_MODEL_PATH.with_suffix('.zip')

VOSK_DOWNLOAD_NOTIFICATION_ID = 'speech_recognition:download-vosk'

OFFLINE_ENGINES: list[SpeechRecognitionEngineName] = [
    SpeechRecognitionEngineName.VOSK,
]



"""Asynchronous evicting queue implementation."""

from __future__ import annotations

import asyncio
from collections import deque
from typing import Generic, TypeVar

T = TypeVar('T')


class AsyncEvictingQueue(Generic[T]):
    """An asynchronous queue that evicts the oldest item when full."""

    def __init__(self, maxsize: int) -> None:
        """Initialize the queue with a maximum size."""
        self.maxsize = maxsize
        self.queue: deque[T] = deque(maxlen=maxsize)
        self._not_empty = asyncio.Condition()

    async def put(self, item: T) -> None:
        """Put an item into the queue, discarding the oldest if full."""
        async with self._not_empty:
            self.queue.append(item)
            self._not_empty.notify()

    async def get(self) -> T:
        """Get an item from the queue, waiting if necessary."""
        async with self._not_empty:
            while True:
                try:
                    return self.queue.popleft()
                except IndexError:
                    await self._not_empty.wait()

    def get_nowait(self) -> T:
        """Get an item from the queue without waiting."""
        return self.queue.popleft()

    def qsize(self) -> int:
        """Return the current size of the queue."""
        return len(self.queue)

    def empty(self) -> bool:
        """Check if the queue is empty."""
        return len(self.queue) == 0

    def full(self) -> bool:
        """Check if the queue is full."""
        return len(self.queue) >= self.maxsize



# pyright: reportMissingModuleSource=false
# ruff: noqa: D100, D101, D103, D107
from __future__ import annotations

import asyncio
import math
import time
from pathlib import Path
from threading import Lock
from typing import TYPE_CHECKING

import headless_kivy.config
import numpy as np
import png
from debouncer import DebounceOptions, debounce
from kivy.clock import Clock

from ubo_app.logger import logger
from ubo_app.store.core.types import (
    CloseApplicationAction,
    OpenApplicationAction,
    RegisterSettingAppAction,
    SettingsCategory,
)
from ubo_app.store.main import store
from ubo_app.store.services.camera import (
    CameraDetectEvent,
    CameraInstallDriverEvent,
    CameraReportBarcodeAction,
    CameraReportImageEvent,
    CameraRestoreDefaultEvent,
    CameraSetAvailableCamerasAction,
    CameraStartViewfinderEvent,
    CameraStopViewfinderEvent,
)
from ubo_app.store.services.display import DisplayPauseAction, DisplayResumeAction
from ubo_app.store.ubo_actions import register_application
from ubo_app.utils import IS_RPI
from ubo_app.utils.async_ import create_task
from ubo_app.utils.error_handlers import report_service_error
from ubo_app.utils.gui import UboPageWidget
from ubo_app.utils.persistent_store import register_persistent_store
from ubo_app.utils.server import send_command

if TYPE_CHECKING:
    from numpy._typing._array_like import NDArray

    from ubo_app.utils.types import Subscriptions

    from .camera_backend import CameraBackend

THROTTL_TIME = 0.5


def resize_image(
    image: NDArray[np.uint8],
    *,
    new_size: tuple[int, int],
) -> NDArray[np.uint8]:
    scale_x = max(image.shape[1] / new_size[1], 1)
    scale_y = max(image.shape[0] / new_size[0], 1)

    # Use slicing to downsample the image
    resized = image[:: int(scale_y), :: int(scale_x)]

    # Handle any rounding issues by trimming the excess
    return resized[: new_size[0], : new_size[1]]


@debounce(
    wait=THROTTL_TIME,
    options=DebounceOptions(leading=True, trailing=False, time_window=THROTTL_TIME),
)
def check_codes(codes: list[str]) -> None:
    store.dispatch(CameraReportBarcodeAction(codes=codes))


class CameraApplication(UboPageWidget):
    def __init__(
        self,
        **kwargs: object,
    ) -> None:
        super().__init__(**kwargs)
        camera = None
        is_running = True

        @store.autorun(lambda state: state.camera.selected_camera_index)
        def _handle_camera_change(index: int) -> None:
            nonlocal camera
            if not is_running:
                return
            # Close existing camera if any
            if camera:
                camera.stop()
                camera.close()
            # Initialize new camera with current index
            camera = initialize_camera(index)

        fs_lock = Lock()

        def feed_viewfinder_locked(_: object) -> None:
            with fs_lock:
                if not is_running:
                    return
                feed_viewfinder(camera)

        feed_viewfinder_scheduler = Clock.schedule_interval(
            feed_viewfinder_locked,
            0.04,
        )

        store.dispatch(DisplayPauseAction())

        def handle_stop_viewfinder(_: object = None) -> None:
            unsubscribe()
            with fs_lock:
                nonlocal is_running
                is_running = False
                feed_viewfinder_scheduler.cancel()
                store.dispatch(
                    CloseApplicationAction(application_instance_id=self.id),
                    DisplayResumeAction(),
                )
                if camera:
                    camera.stop()
                    camera.close()

        self.bind(on_close=handle_stop_viewfinder)

        unsubscribe = store.subscribe_event(
            CameraStopViewfinderEvent,
            handle_stop_viewfinder,
        )


register_application(application_id='camera:viewfinder', application=CameraApplication)


def initialize_camera(camera_index: int = 0) -> CameraBackend | None:
    """Initialize the appropriate camera backend based on platform.

    Args:
        camera_index: Camera device index (default: 0, only used on non-RPI platforms)

    Returns:
        Camera backend instance or None if initialization fails

    """
    try:
        width = headless_kivy.config.width() * 2
        height = headless_kivy.config.height() * 2

        if IS_RPI:
            from picamera2_backend import PiCamera2Backend

            logger.info(
                'Initializing camera with index {index}',
                extra={'index': camera_index},
            )
            camera = PiCamera2Backend(
                width=width,
                height=height,
                camera_index=camera_index,
            )
        else:
            from opencv_backend import OpenCVCameraBackend

            logger.info(
                'Initializing camera with index {index}',
                extra={'index': camera_index},
            )
            camera = OpenCVCameraBackend(
                width=width,
                height=height,
                camera_index=camera_index,
            )

        camera.start()
    except Exception:
        report_service_error()
        logger.exception('Failed to initialize camera.')
        return None
    else:
        return camera



def feed_viewfinder(camera: CameraBackend | None) -> None:
    width = headless_kivy.config.width()
    height = headless_kivy.config.height()

    if not IS_RPI:
        path = Path('/tmp/qrcode_input.txt')  # noqa: S108
        if path.exists():
            barcodes = [path.read_text().strip()]
            path.unlink(missing_ok=True)
            create_task(check_codes(codes=barcodes))
            return

    qrcode_path = Path('/tmp/qrcode_input.png')  # noqa: S108
    if qrcode_path.exists():
        with qrcode_path.open('rb') as file:
            reader = png.Reader(file)
            width, height, data, _ = reader.read()
            data = np.array(list(data)).reshape((height, width, 4))
        qrcode_path.unlink(missing_ok=True)
    elif camera:
        data = camera.capture_array('main')
    else:
        data = None

    if data is not None:
        from pyzbar.pyzbar import decode

        barcodes = decode(data)
        create_task(
            check_codes(codes=[barcode.data.decode() for barcode in barcodes]),
        )

        data = resize_image(data, new_size=(width, height))

        # Mirror the image
        data = np.rot90(data, 2)[:, ::-1, :3]

        viewfinder_data = data.astype(np.uint16)

        # Render an empty rounded rectangle
        margin = 15
        thickness = 7

        lines = [
            ((margin, width - margin), (margin, margin + thickness)),
            (
                (margin, width - margin),
                (height - margin - thickness, height - margin),
            ),
            (
                (margin, margin + thickness),
                (margin + thickness, height - margin - thickness),
            ),
            (
                (width - margin - thickness, width - margin),
                (margin + thickness, height - margin - thickness),
            ),
        ]
        for line in lines:
            viewfinder_data[line[0][0] : line[0][1], line[1][0] : line[1][1]] = (
                0xFF - viewfinder_data[line[0][0] : line[0][1], line[1][0] : line[1][1]]
            ) // 2

        color = (
            (viewfinder_data[:, :, 2] & 0xF8) << 8
            | (viewfinder_data[:, :, 1] & 0xFC) << 3
            | viewfinder_data[:, :, 0] >> 3
        )

        data_bytes = bytes(
            np.dstack(((color >> 8) & 0xFF, color & 0xFF)).flatten().tolist(),
        )

        from ubo_app.display import display

        display.render_block(
            rectangle=(0, 0, width - 1, height - 1),
            data_bytes=data_bytes,
            bypass_pause=True,
        )

        store._dispatch(  # noqa: SLF001
            [
                CameraReportImageEvent(
                    timestamp=time.time(),
                    data=data.tobytes(),
                    width=width,
                    height=height,
                ),
            ],
        )


async def _install_camera_driver(event: CameraInstallDriverEvent) -> None:
    from ubo_app.colors import DANGER_COLOR, SUCCESS_COLOR, WARNING_COLOR
    from ubo_app.store.core.types import RebootAction
    from ubo_app.store.services.notifications import (
        Chime,
        Notification,
        NotificationDispatchItem,
        NotificationDisplayType,
        NotificationsAddAction,
    )

    store.dispatch(
        NotificationsAddAction(
            notification=Notification(
                id='camera_install_driver',
                title='Camera Driver',
                content=f'Installing {event.make} {event.model} driver...',
                display_type=NotificationDisplayType.STICKY,
                color=WARNING_COLOR,
                icon='󰄁',
                show_dismiss_action=False,
                progress=math.nan,
            ),
        ),
    )
    result = await send_command(
        'camera',
        'install_driver',
        event.make,
        event.model,
        event.variant,
        has_output=True,
    )
    if result == 'installed':
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id='camera_install_driver',
                    title='Camera Driver',
                    content=f'{event.make} {event.model} installed successfully.\n'
                    'Reboot required for changes to take effect.',
                    display_type=NotificationDisplayType.STICKY,
                    color=SUCCESS_COLOR,
                    icon='󰄬',
                    chime=Chime.DONE,
                    show_dismiss_action=True,
                    actions=[
                        NotificationDispatchItem(
                            icon='󰜉',
                            store_action=RebootAction(),
                        ),
                    ],
                ),
            ),
        )
    else:
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id='camera_install_driver',
                    title='Camera Driver',
                    content=f'Failed to install {event.make} {event.model} driver',
                    display_type=NotificationDisplayType.FLASH,
                    color=DANGER_COLOR,
                    icon='󰜺',
                    chime=Chime.FAILURE,
                ),
            ),
        )


async def _restore_default_camera(_: CameraRestoreDefaultEvent) -> None:
    from ubo_app.colors import DANGER_COLOR, SUCCESS_COLOR, WARNING_COLOR
    from ubo_app.store.core.types import RebootAction
    from ubo_app.store.services.notifications import (
        Chime,
        Notification,
        NotificationDispatchItem,
        NotificationDisplayType,
        NotificationsAddAction,
    )

    store.dispatch(
        NotificationsAddAction(
            notification=Notification(
                id='camera_restore_default',
                title='Camera',
                content='Restoring default camera configuration...',
                display_type=NotificationDisplayType.STICKY,
                color=WARNING_COLOR,
                icon='󰁯',
                show_dismiss_action=False,
                progress=math.nan,
            ),
        ),
    )
    # Allow the UI to fully render the progress notification before proceeding.
    # Without this delay, the operation completes so fast that the replacement
    # notification is dispatched before the STICKY view finishes rendering,
    # causing the progress notification to stay visible instead of updating.
    await asyncio.sleep(0.5)
    result = await send_command(
        'camera',
        'restore_default',
        has_output=True,
    )
    if result == 'restored':
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id='camera_restore_default',
                    title='Camera',
                    content='Default camera restored successfully.\n'
                    'Reboot required for changes to take effect.',
                    display_type=NotificationDisplayType.STICKY,
                    color=SUCCESS_COLOR,
                    icon='󰄬',
                    chime=Chime.DONE,
                    show_dismiss_action=True,
                    actions=[
                        NotificationDispatchItem(
                            icon='󰜉',
                            store_action=RebootAction(),
                        ),
                    ],
                ),
            ),
        )
    else:
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id='camera_restore_default',
                    title='Camera',
                    content='Failed to restore default camera configuration',
                    display_type=NotificationDisplayType.FLASH,
                    color=DANGER_COLOR,
                    icon='󰜺',
                    chime=Chime.FAILURE,
                ),
            ),
        )


def start_camera_viewfinder() -> None:
    store.dispatch(
        OpenApplicationAction(application_id='camera:viewfinder'),
    )


async def detect_and_update_cameras() -> None:
    """Detect available cameras and update state."""
    try:
        if IS_RPI:
            from utils import detect_available_cameras_picamera2

            logger.info('Starting Picamera2 camera detection...')
            available = detect_available_cameras_picamera2()
        else:
            from utils import detect_available_cameras

            logger.info('Starting OpenCV camera detection...')
            available = detect_available_cameras()

        logger.info(
            'Camera detection complete: {count} camera(s) found',
            extra={'count': len(available), 'indices': available},
        )
        store.dispatch(CameraSetAvailableCamerasAction(available_cameras=available))
    except Exception:
        logger.exception('Error during camera detection')
        store.dispatch(CameraSetAvailableCamerasAction(available_cameras=[]))


def handle_camera_detect(_: CameraDetectEvent) -> None:
    """Handle camera detection event."""
    logger.info('Camera detect event received, starting detection...')
    create_task(detect_and_update_cameras())


def init_service() -> Subscriptions:
    from pages import CameraSettingsMenu

    # Register camera settings menu
    store.dispatch(
        RegisterSettingAppAction(
            priority=1,
            category=SettingsCategory.HARDWARE,
            menu_item=CameraSettingsMenu,
        ),
    )

    # Detect cameras on startup
    create_task(detect_and_update_cameras())

    # Register persistent storage for selected camera index
    register_persistent_store(
        'camera_selected_index',
        lambda state: state.camera.selected_camera_index,
    )

    # Register persistent storage for camera type
    register_persistent_store(
        'camera_type',
        lambda state: state.camera.camera_type,
    )

    return [
        store.subscribe_event(
            CameraStartViewfinderEvent,
            start_camera_viewfinder,
        ),
        store.subscribe_event(
            CameraDetectEvent,
            handle_camera_detect,
        ),
        store.subscribe_event(
            CameraInstallDriverEvent,
            _install_camera_driver,
        ),
        store.subscribe_event(
            CameraRestoreDefaultEvent,
            _restore_default_camera,
        ),
    ]



# ruff: noqa: D100, D103
from __future__ import annotations

from dataclasses import replace
from typing import TYPE_CHECKING

from engines_registry import (
    IMAGE_GENERATOR_ENGINES,
    LLM_ENGINES,
    STT_ENGINES,
    TTS_ENGINES,
)
from redux import CompleteReducerResult, InitializationActionError
from redux.basic_types import InitAction

from ubo_app.logger import logger
from ubo_app.store.services.assistant import (
    AssistantAction,
    AssistantAddMcpServerAction,
    AssistantAddMcpServerEvent,
    AssistantDeleteMcpServerAction,
    AssistantDeleteMcpServerEvent,
    AssistantDownloadOllamaModelAction,
    AssistantDownloadOllamaModelEvent,
    AssistantEvent,
    AssistantHandleReportEvent,
    AssistantReportAction,
    AssistantSetIsActiveAction,
    AssistantSetSelectedImageGeneratorAction,
    AssistantSetSelectedLLMAction,
    AssistantSetSelectedModelAction,
    AssistantSetSelectedSTTAction,
    AssistantSetSelectedTTSAction,
    AssistantStartListeningAction,
    AssistantState,
    AssistantStopListeningAction,
    AssistantSyncMcpServersAction,
    AssistantToggleListeningAction,
    AssistantToggleMcpServerAction,
    AssistantUpdateProvidersAction,
)
from ubo_app.store.services.audio import (
    AudioAction,
    AudioDevice,
    AudioSetMuteStatusAction,
)
from ubo_app.store.services.notifications import (
    Chime,
    Importance,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.rgb_ring import RgbRingBlankAction, RgbRingRainbowAction

if TYPE_CHECKING:
    from redux import ReducerResult

    from ubo_app.store.services.notifications import NotificationsAction
    from ubo_app.store.services.rgb_ring import RgbRingAction

# MCP server_id format: name_uuid (2 parts when split by last underscore)
_MCP_SERVER_ID_PARTS = 2


def reducer(
    state: AssistantState | None,
    action: AssistantAction | AudioAction,
) -> ReducerResult[AssistantState, RgbRingAction | NotificationsAction, AssistantEvent]:
    if state is None:
        if isinstance(action, InitAction):
            return AssistantState()

        raise InitializationActionError(action)

    match action:
        case AssistantSetIsActiveAction():
            return replace(state, is_active=action.is_active)

        case AssistantSetSelectedSTTAction():
            return replace(state, selected_stt=action.stt_name)

        case AssistantSetSelectedLLMAction():
            return replace(state, selected_llm=action.llm_name)

        case AssistantSetSelectedTTSAction():
            return replace(state, selected_tts=action.tts_name)

        case AssistantSetSelectedImageGeneratorAction():
            return replace(state, selected_image_generator=action.image_generator_name)

        case AssistantSetSelectedModelAction():
            return replace(
                state,
                selected_models={
                    **state.selected_models,
                    state.selected_llm: action.model,
                },
            )

        case AssistantDownloadOllamaModelAction():
            return CompleteReducerResult(
                state=state,
                events=[AssistantDownloadOllamaModelEvent(model=action.model)],
            )

        case AssistantUpdateProvidersAction():
            all_engines = {
                **STT_ENGINES,
                **TTS_ENGINES,
                **LLM_ENGINES,
                **IMAGE_GENERATOR_ENGINES,
            }
            # Build setup status dict - this is the source of truth
            # Use getattr since not all engines have is_setup (only NeedsSetupMixin)
            provider_setup_status = {
                engine.name: getattr(engine, 'is_setup', True)
                for engine in all_engines.values()
            }
            return replace(
                state,
                provider_setup_status=provider_setup_status,
            )

        case AssistantReportAction():
            return CompleteReducerResult(
                state=state,
                events=[
                    AssistantHandleReportEvent(
                        source_id=action.source_id,
                        data=action.data,
                    ),
                ],
            )

        case AudioSetMuteStatusAction(device=AudioDevice.INPUT):
            return replace(state, is_microphone_mute=action.is_mute)

        case AssistantStartListeningAction():
            if state.is_microphone_mute:
                return CompleteReducerResult(
                    state=state,
                    actions=[
                        NotificationsAddAction(
                            notification=Notification(
                                title='Microphone Muted',
                                content='Microphone is mute. Unmute to speak.',
                                importance=Importance.HIGH,
                                icon='󰍭',
                                display_type=NotificationDisplayType.STICKY,
                                chime=Chime.FAILURE,
                            ),
                        ),
                    ],
                )
            return CompleteReducerResult(
                state=state(is_listening=True),
                actions=[RgbRingRainbowAction(rounds=0, wait=800)],
            )

        case AssistantStopListeningAction():
            return CompleteReducerResult(
                state=state(is_listening=False),
                actions=[RgbRingBlankAction()],
            )

        case AssistantToggleListeningAction():
            if state.is_listening:
                # Currently listening, stop it
                return CompleteReducerResult(
                    state=state(is_listening=False),
                    actions=[RgbRingBlankAction()],
                )
            # Not listening, start it (with mute check)
            if state.is_microphone_mute:
                return CompleteReducerResult(
                    state=state,
                    actions=[
                        NotificationsAddAction(
                            notification=Notification(
                                title='Microphone Muted',
                                content='Microphone is mute. Unmute to speak.',
                                importance=Importance.HIGH,
                                icon='󰍭',
                                display_type=NotificationDisplayType.STICKY,
                                chime=Chime.FAILURE,
                            ),
                        ),
                    ],
                )
            return CompleteReducerResult(
                state=state(is_listening=True),
                actions=[RgbRingRainbowAction(rounds=0, wait=800)],
            )

        case AssistantAddMcpServerAction():
            logger.info(
                'AssistantAddMcpServerAction received',
                extra={'server_name': action.name, 'mcp_type': action.type.value},
            )
            return CompleteReducerResult(
                state=state,
                events=[
                    AssistantAddMcpServerEvent(
                        name=action.name,
                        type=action.type,
                        config=action.config,
                    ),
                ],
            )

        case AssistantToggleMcpServerAction():
            from mcp_servers import toggle_mcp_server

            # Toggle in filesystem
            new_state = toggle_mcp_server(action.server_id)
            logger.info(
                'AssistantToggleMCPServerAction processed',
                extra={'server_id': action.server_id, 'enabled': new_state},
            )

            # Update in-memory state
            enabled_servers = list(state.enabled_mcp_servers)
            if new_state:
                if action.server_id not in enabled_servers:
                    enabled_servers.append(action.server_id)
            elif action.server_id in enabled_servers:
                    enabled_servers.remove(action.server_id)

            # Serialize enabled servers with metadata for gRPC autorun
            import json

            enabled_with_metadata = [
                {
                    'server_id': state.mcp_servers[sid].server_id,
                    'name': state.mcp_servers[sid].name,
                    'type': state.mcp_servers[sid].type.value,  # Convert enum to string
                    'config': state.mcp_servers[sid].config,
                }
                for sid in enabled_servers
                if sid in state.mcp_servers
            ]
            enabled_mcp_servers_with_metadata_json = json.dumps(enabled_with_metadata)

            return replace(
                state,
                enabled_mcp_servers=enabled_servers,
                enabled_mcp_servers_with_metadata_json=enabled_mcp_servers_with_metadata_json,
            )

        case AssistantDeleteMcpServerAction():
            # Remove from enabled servers if present
            enabled_servers = list(state.enabled_mcp_servers)
            if action.server_id in enabled_servers:
                enabled_servers.remove(action.server_id)
            # Remove from mcp_servers dict
            mcp_servers = {
                k: v for k, v in state.mcp_servers.items() if k != action.server_id
            }
            # Serialize enabled servers with metadata for gRPC autorun
            import json

            enabled_with_metadata = [
                {
                    'server_id': mcp_servers[sid].server_id,
                    'name': mcp_servers[sid].name,
                    'type': mcp_servers[sid].type.value,  # Convert enum to string
                    'config': mcp_servers[sid].config,
                }
                for sid in enabled_servers
                if sid in mcp_servers
            ]
            enabled_mcp_servers_with_metadata_json = json.dumps(enabled_with_metadata)

            logger.info(
                'AssistantDeleteMCPServerAction processed',
                extra={
                    'server_id': action.server_id,
                    'remaining_servers': len(mcp_servers),
                    'remaining_enabled': len(enabled_servers),
                },
            )
            return CompleteReducerResult(
                state=replace(
                    state,
                    enabled_mcp_servers=enabled_servers,
                    mcp_servers=mcp_servers,
                    enabled_mcp_servers_with_metadata_json=enabled_mcp_servers_with_metadata_json,
                ),
                events=[AssistantDeleteMcpServerEvent(server_id=action.server_id)],
            )

        case AssistantSyncMcpServersAction():
            # Load servers from filesystem and update state
            import json

            from ubo_app.constants.assistant import ASSISTANT_MCP_SERVERS_PATH
            from ubo_app.store.services.assistant import (
                McpServerMetadata,
                McpServerType,
            )

            loaded_servers: dict[str, McpServerMetadata] = {}
            enabled_servers: list[str] = []

            logger.debug(
                'Syncing MCP servers from filesystem',
                extra={'path': str(ASSISTANT_MCP_SERVERS_PATH),
                        'exists': ASSISTANT_MCP_SERVERS_PATH.exists(),
                    },
            )

            if ASSISTANT_MCP_SERVERS_PATH.exists():
                # Iterate through server directories
                for server_dir in ASSISTANT_MCP_SERVERS_PATH.iterdir():
                    if not server_dir.is_dir():
                        continue

                    config_file = server_dir / 'config.json'
                    if not config_file.exists():
                        continue

                    try:
                        with config_file.open() as f:
                            data = json.load(f)

                        server_id = server_dir.name
                        # Extract name from server_id (format: name_uuid)
                        name_parts = server_id.rsplit('_', 1)
                        name = (
                            name_parts[0]
                            if len(name_parts) == _MCP_SERVER_ID_PARTS
                            else server_id
                        )

                        # Ensure config is always a string
                        # (JSON for dicts, plain string for URLs)
                        config = data['config']
                        if isinstance(config, dict):
                            config = json.dumps(config)

                        loaded_servers[server_id] = McpServerMetadata(
                            server_id=server_id,
                            name=name,
                            type=McpServerType(data['type']),  # Convert string to enum
                            config=config,
                        )

                        # Track enabled state from config file
                        if data.get('enabled', False):
                            enabled_servers.append(server_id)

                        logger.debug(
                            'Loaded MCP server',
                            extra={
                                'server_id': server_id,
                                'server_name': name,
                                'enabled': data.get('enabled', False),
                            },
                        )
                    except Exception:
                        logger.exception(
                            'Failed to load MCP server',
                            extra={'config_file': str(config_file)},
                        )
                        continue

            # Serialize enabled servers with metadata for gRPC autorun
            enabled_with_metadata = [
                {
                    'server_id': loaded_servers[sid].server_id,
                    'name': loaded_servers[sid].name,
                    'type': loaded_servers[sid].type.value,  # Convert enum to string
                    'config': loaded_servers[sid].config,
                }
                for sid in enabled_servers
                if sid in loaded_servers
            ]
            enabled_mcp_servers_with_metadata_json = json.dumps(enabled_with_metadata)

            logger.debug(
                'Finished syncing MCP servers',
                extra={'server_count': len(loaded_servers),
                    'server_ids': list(loaded_servers.keys()),
                    'enabled_count': len(enabled_servers),
                    },
            )

            return replace(
                state,
                mcp_servers=loaded_servers,
                enabled_mcp_servers=enabled_servers,
                enabled_mcp_servers_with_metadata_json=enabled_mcp_servers_with_metadata_json,
            )

        case _:
            return state



"""Implementation of the web-ui service."""

import asyncio
import datetime
import functools
import json
import re
from pathlib import Path
from typing import TYPE_CHECKING, Literal, cast

from quart import Quart, Response, render_template, request
from ubo_bindings.ubo.v1 import WebUiState as GRPCWebUIState

from ubo_app.constants import (
    GRPC_ENVOY_LISTEN_PORT,
    WEB_UI_DEBUG_MODE,
    WEB_UI_HOTSPOT_PASSWORD,
    WEB_UI_LISTEN_ADDRESS,
    WEB_UI_LISTEN_PORT,
)
from ubo_app.logger import logger
from ubo_app.rpc.object_to_message import build_message
from ubo_app.store.input.types import (
    InputCancelAction,
    InputMethod,
    InputProvideAction,
    InputResult,
)
from ubo_app.store.main import store
from ubo_app.store.services.docker import (
    DockerImageFetchAction,
    DockerImageRemoveAction,
    DockerImageRemoveContainerAction,
    DockerImageRunAction,
    DockerImageStopAction,
    DockerInstallAction,
    DockerStartAction,
    DockerStopAction,
)
from ubo_app.store.services.notifications import (
    Importance,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.speech_synthesis import ReadableInformation
from ubo_app.store.services.web_ui import (
    WebUIInitializeEvent,
    WebUIState,
    WebUIStopEvent,
)
from ubo_app.utils.async_ import create_task
from ubo_app.utils.error_handlers import report_service_error
from ubo_app.utils.network import has_gateway
from ubo_app.utils.pod_id import get_pod_id
from ubo_app.utils.server import send_command
from ubo_app.utils.types import Subscriptions

if TYPE_CHECKING:
    from werkzeug.datastructures import FileStorage

ENVOY_IMAGE_NAME = 'thegrandpkizzle/envoy:1.26.1'


async def _get_docker_status() -> str:
    try:
        process = await asyncio.subprocess.create_subprocess_exec(
            'docker',
            'info',
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        await asyncio.wait_for(process.wait(), timeout=2)
        if process.returncode is None:
            process.kill()
        if process.stdout and process.returncode == 0:
            output = await process.stdout.read()
            return 'running' if 'Containers' in output.decode() else 'not ready'
    except FileNotFoundError:
        logger.warning('Docker is not installed')
        return 'not installed'
    except Exception:
        logger.exception('Failed to check if docker is running')
        report_service_error()
        return 'failed'
    else:
        logger.warning('Docker process returned non-zero exit code')
        return 'not running'


async def _get_envoy_status() -> str:
    try:
        process = await asyncio.subprocess.create_subprocess_exec(
            'docker',
            'inspect',
            ENVOY_IMAGE_NAME,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        await asyncio.wait_for(process.wait(), timeout=2)
        if process.returncode is None:
            process.kill()
        if process.stdout and process.returncode == 0:
            output = await process.stdout.read()
            if ENVOY_IMAGE_NAME in output.decode():
                process = await asyncio.subprocess.create_subprocess_exec(
                    'docker',
                    'ps',
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE,
                )
                await asyncio.wait_for(process.wait(), timeout=3)
                if process.returncode is None:
                    process.kill()
                if process.stdout and process.returncode == 0:
                    output = await process.stdout.read()
                    return (
                        'running'
                        if ENVOY_IMAGE_NAME in output.decode()
                        else 'not running'
                    )

                logger.warning('Docker process returned non-zero exit code')
                return 'not running'

            return 'not running'
        else:  # noqa: RET505
            logger.warning('Docker process returned non-zero exit code')
            return 'not downloaded'
    except Exception:
        logger.exception('Failed to check if envoy is running')
        report_service_error()
        return 'failed'


async def initialize(event: WebUIInitializeEvent) -> None:
    """Start the hotspot if there is no network connection."""
    is_connected = await has_gateway()
    logger.info(
        'web-ui - initialize hotspot',
        extra={
            'is_connected': is_connected,
            'description': event.description,
        },
    )
    if not is_connected:
        result = await send_command('hotspot', 'start', has_output=True)
        if result != 'done':
            store.dispatch(
                InputCancelAction(id=event.description.id),
                NotificationsAddAction(
                    notification=Notification(
                        id='web_ui:hotspot_error',
                        icon='󱋆',
                        title='Web UI Error',
                        content='Failed to start the hotspot, please check the logs.',
                        display_type=NotificationDisplayType.STICKY,
                        importance=Importance.HIGH,
                    ),
                ),
            )
            return
    store.dispatch(
        NotificationsAddAction(
            notification=Notification(
                id=f'web_ui:pending:{event.description.id}',
                icon='󱋆',
                title='Web UI',
                content=f'[size=18dp]{event.description.prompt}[/size]',
                display_type=NotificationDisplayType.STICKY,
                is_read=True,
                extra_information=ReadableInformation(
                    text=(
                        'Please make sure you are on the same network as this '
                        'ubo-pod and open '
                        f'http://{{{{hostname}}}}:{WEB_UI_LISTEN_PORT} in your browser.'
                        if is_connected
                        else f'Please connect to the "{get_pod_id()}" WiFi network '
                        f'with password "{WEB_UI_HOTSPOT_PASSWORD}" and open '
                        f'http://{{{{hostname}}}}:{WEB_UI_LISTEN_PORT} in your browser.'
                    ),
                ),
                expiration_timestamp=datetime.datetime.now(tz=datetime.UTC),
                show_dismiss_action=False,
                dismiss_on_close=True,
                on_close=functools.partial(
                    store.dispatch,
                    InputCancelAction(id=event.description.id),
                ),
            ),
        ),
    )


async def stop_hotspot(*, silence: bool = False) -> None:
    """Start the hotspot if there is no network connection."""
    logger.info('web-ui - stop hotspot')
    result = await send_command('hotspot', 'stop', has_output=True)
    if result != 'done' and not silence:
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id='web_ui:hotspot_error',
                    icon='󱋆',
                    title='Web UI Error',
                    content='Failed to stop the hotspot properly, '
                    'please check the logs.',
                    display_type=NotificationDisplayType.STICKY,
                    importance=Importance.HIGH,
                ),
            ),
        )


async def init_service() -> Subscriptions:  # noqa: C901, PLR0915
    """Initialize the web-ui service."""
    _ = []
    app = Quart(
        'ubo-app',
        template_folder=(Path(__file__).parent / 'templates').absolute().as_posix(),
        static_folder=(Path(__file__).parent / 'web-app' / 'dist')
        .absolute()
        .as_posix(),
    )
    app.debug = WEB_UI_DEBUG_MODE
    shutdown_event: asyncio.Event = asyncio.Event()

    @store.with_state(lambda state: state.web_ui)
    def state(state: WebUIState) -> str:
        return (
            build_message(state, expected_type=GRPCWebUIState).SerializeToString().hex()
        )

    @app.route('/', methods=['GET', 'POST'])
    async def inputs_form() -> str:
        if request.method == 'POST':
            data: dict[str, str] = dict(await request.form)
            files = {
                key: cast('FileStorage', value).stream.read()
                for key, value in (await request.files).items()
            }

            if data['action'] == 'cancel':
                store.dispatch(InputCancelAction(id=data['id']))
            elif data['action'] == 'provide':
                id = data.pop('id')
                value = data.pop('value', '')
                store.dispatch(
                    InputProvideAction(
                        id=id,
                        value=value,
                        result=InputResult(
                            data=data,
                            files=files,
                            method=InputMethod.WEB_DASHBOARD,
                        ),
                    ),
                )
            await asyncio.sleep(0.1)
        return await render_template(
            'index.jinja2',
            state=state(),
            re=re,
            GRPC_ENVOY_LISTEN_PORT=GRPC_ENVOY_LISTEN_PORT,
        )

    @app.route('/status')
    async def status() -> Response:
        statuses = await asyncio.gather(
            _get_docker_status(),
            _get_envoy_status(),
        )
        return Response(
            json.dumps(
                {
                    'status': 'ok',
                    'docker': statuses[0],
                    'envoy': statuses[1],
                    'state': state(),
                },
            ),
            content_type='application/json',
        )

    @app.route('/action/', methods=['POST'])
    async def action() -> Response:
        data = await request.json
        action: Literal[
            'install docker',
            'run docker',
            'stop docker',
            'download envoy',
            'run envoy',
            'remove envoy',
        ] = data['action']
        if action == 'install docker':
            store.dispatch(DockerInstallAction())
        elif action == 'run docker':
            store.dispatch(DockerStartAction())
        elif action == 'stop docker':
            store.dispatch(DockerStopAction())
        elif action == 'download envoy':
            store.dispatch(DockerImageFetchAction(image='envoy_grpc'))
        elif action == 'run envoy':
            store.dispatch(DockerImageRunAction(image='envoy_grpc'))
        elif action == 'remove envoy':
            store.dispatch(DockerImageStopAction(image='envoy_grpc'))
            await asyncio.sleep(2)
            store.dispatch(DockerImageRemoveContainerAction(image='envoy_grpc'))
            await asyncio.sleep(2)
            store.dispatch(DockerImageRemoveAction(image='envoy_grpc'))
        return Response(
            json.dumps({'status': 'ok'}),
            content_type='application/json',
        )

    if WEB_UI_DEBUG_MODE:

        @app.errorhandler(Exception)
        async def handle_error(_: Exception) -> str:
            import traceback

            return f'<pre>{traceback.format_exc()}</pre>'

        _.append(handle_error)

    store.subscribe_event(WebUIInitializeEvent, initialize)
    store.subscribe_event(WebUIStopEvent, stop_hotspot)

    start_event = asyncio.Event()

    async def wait_for_shutdown() -> None:
        await shutdown_event.wait()

    app.before_serving(start_event.set)

    create_task(
        app.run_task(
            host=WEB_UI_LISTEN_ADDRESS,
            port=WEB_UI_LISTEN_PORT,
            debug=WEB_UI_DEBUG_MODE,
            shutdown_trigger=wait_for_shutdown,
        ),
    )

    await start_event.wait()

    async def cleanup() -> None:
        shutdown_event.set()
        await stop_hotspot(silence=True)

    return [cleanup]



"""Some useful type definitions."""

from __future__ import annotations

from collections.abc import Callable, Coroutine, Sequence
from typing import TYPE_CHECKING, Protocol, TypeAlias

from typing_extensions import TypeVar

if TYPE_CHECKING:
    from asyncio import Handle

    from redux.basic_types import TaskCreatorCallback

Subscriptions: TypeAlias = Sequence[Callable[[], None | Coroutine[None, None, None]]]


TaskType = TypeVar('TaskType', infer_variance=True)


class CoroutineRunner(Protocol):
    """Run a coroutine in the event loop."""

    def __call__(
        self,
        coroutine: Coroutine[None, None, TaskType],
        callback: TaskCreatorCallback | None = None,
        name: str | None = None,
    ) -> Handle:
        """Run a coroutine in the event loop.

        Args:
            coroutine: The coroutine to run.
            callback: A callback to call when the coroutine is done.
            name: An optional name for the task.

        Returns:
            The handle of the task.

        """
        ...



"""Vosk engine interface."""

from __future__ import annotations

import asyncio
import shutil

from typing_extensions import override
from ubo_gui.constants import WARNING_COLOR

from ubo_app.colors import DANGER_COLOR, INFO_COLOR
from ubo_app.constants.assistant import (
    VOSK_DOWNLOAD_NOTIFICATION_ID,
    VOSK_DOWNLOAD_PATH,
    VOSK_MODEL_PATH,
    VOSK_MODEL_URL,
)
from ubo_app.engines.abstraction.ai_provider_mixin import AIProviderMixin
from ubo_app.engines.abstraction.needs_setup_mixin import NeedsSetupMixin
from ubo_app.store.main import store
from ubo_app.store.services.assistant import AssistantUpdateProvidersAction
from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationActionItem,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.speech_recognition import (
    SpeechRecognitionSetIsIntentsActiveAction,
)
from ubo_app.store.services.speech_synthesis import ReadableInformation
from ubo_app.utils.async_ import create_task
from ubo_app.utils.download import download_file


class VoskEngine(NeedsSetupMixin, AIProviderMixin):
    """Vosk engine."""

    @property
    def name(self) -> str:
        """The internal name of the Vosk engine."""
        return 'vosk'

    @property
    def label(self) -> str:
        """The display label for the Vosk engine."""
        return 'Vosk'

    @property
    def not_setup_message(self) -> str:
        """Message shown when the Vosk service API key is not set."""
        return 'Vosk model path does not exist. Please download it in the settings.'

    def _update_download_notification(self, *, progress: float) -> None:
        extra_information = ReadableInformation(
            text="""\
The download progress is shown in the radial progress bar at the top left corner of \
the screen.""",
        )
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id=VOSK_DOWNLOAD_NOTIFICATION_ID,
                    title='Downloading',
                    content='Vosk speech recognition model',
                    extra_information=extra_information,
                    display_type=NotificationDisplayType.FLASH
                    if progress == 1
                    else NotificationDisplayType.STICKY,
                    flash_time=1,
                    color=INFO_COLOR,
                    icon='󰇚',
                    blink=False,
                    progress=progress,
                    show_dismiss_action=progress == 1,
                    dismiss_on_close=progress == 1,
                ),
            ),
        )

    def _handle_error(self) -> None:
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    id=VOSK_DOWNLOAD_NOTIFICATION_ID,
                    title='Vosk',
                    content='Failed to download',
                    display_type=NotificationDisplayType.STICKY,
                    color=DANGER_COLOR,
                    icon='󰜺',
                    chime=Chime.FAILURE,
                ),
            ),
        )
        shutil.rmtree(VOSK_MODEL_PATH, ignore_errors=True)

    def _download_vosk_model(self) -> None:
        """Download Vosk model."""
        shutil.rmtree(VOSK_MODEL_PATH, ignore_errors=True)

        self._update_download_notification(progress=0)

        async def download() -> None:
            try:
                VOSK_DOWNLOAD_PATH.parent.mkdir(parents=True, exist_ok=True)
                VOSK_MODEL_PATH.parent.mkdir(parents=True, exist_ok=True)

                async for downloaded_bytes, size in download_file(
                    url=VOSK_MODEL_URL,
                    path=VOSK_DOWNLOAD_PATH,
                ):
                    if size:
                        self._update_download_notification(
                            progress=min(1.0, downloaded_bytes / size),
                        )

                self._update_download_notification(progress=1.0)

                process = await asyncio.create_subprocess_exec(
                    '/usr/bin/env',
                    'unzip',
                    '-o',
                    VOSK_DOWNLOAD_PATH,
                    '-d',
                    VOSK_MODEL_PATH.parent,
                    stdout=asyncio.subprocess.DEVNULL,
                    stderr=asyncio.subprocess.DEVNULL,
                )
                await process.wait()
                store.dispatch(
                    SpeechRecognitionSetIsIntentsActiveAction(is_active=True),
                    AssistantUpdateProvidersAction(),
                )
                decide = getattr(self, 'decide_running_state', None)
                if decide is not None:
                    decide()
            except Exception:
                self._handle_error()
                raise
            finally:
                VOSK_DOWNLOAD_PATH.unlink(missing_ok=True)
                self.event.set()

        create_task(download())

    @override
    async def _setup(self) -> None:
        if self.is_setup:
            return
        from ubo_app.store.main import store

        self.event = asyncio.Event()
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    title='Vosk Engine Setup',
                    content='Download the Vosk model.',
                    color=WARNING_COLOR,
                    actions=[
                        NotificationActionItem(
                            label='Download Model',
                            icon='󰇚',
                            action=self._download_vosk_model,
                        ),
                    ],
                ),
            ),
        )
        await self.event.wait()

    @property
    @override
    def is_setup(self) -> bool:
        """Check if the Vosk model is set up."""
        return VOSK_MODEL_PATH.exists()



"""Constants for the assistant module."""

import os

from ubo_app.constants import CONFIG_PATH, DATA_PATH

INTENTS_WAKE_WORD = os.environ.get('UBO_INTENTS_WAKE_WORD', 'short voice command')
ASSISTANT_WAKE_WORD = os.environ.get('UBO_ASSISTANT_WAKE_WORD', 'can you help me')
ASSISTANT_END_WORD = os.environ.get('UBO_ASSISTANT_END_WORD', 'roger that')
ASSISTANT_DEBUG_PATH = os.environ.get('UBO_ASSISTANT_DEBUG_PATH')
DEFAULT_LLM_OLLAMA_MODEL = os.environ.get(
    'UBO_DEFAULT_ASSISTANT_OLLAMA_MODEL',
    'gemma3:1b',
)
DEFAULT_LLM_GOOGLE_MODEL = os.environ.get(
    'UBO_DEFAULT_ASSISTANT_GOOGLE_MODEL',
    'gemini-2.5-flash-preview-05-20',
)
DEFAULT_LLM_OPENAI_MODEL = os.environ.get(
    'UBO_DEFAULT_ASSISTANT_OPENAI_MODEL',
    'gpt-4o',
)
DEFAULT_LLM_GROK_MODEL = os.environ.get(
    'UBO_DEFAULT_ASSISTANT_GROK_MODEL',
    'grok-4-0709',
)
DEFAULT_LLM_CEREBRAS_MODEL = os.environ.get(
    'UBO_DEFAULT_ASSISTANT_CEREBRAS_MODEL',
    'qwen-3-235b-a22b-instruct-2507',
)
DEFAULT_LLM_OLLAMA_ONPREM_MODEL = os.environ.get(
    'UBO_DEFAULT_ASSISTANT_OLLAMA_ONPREM_MODEL',
    'granite3.3:8b',
)

GOOGLE_API_KEY_SECRET_ID = 'google_api_key'  # noqa: S105
GOOGLE_API_KEY_PATTERN = '^AIza[0-9A-Za-z\\-_]{35}$'

GOOGLE_CLOUD_SERVICE_ACCOUNT_KEY_SECRET_ID = 'google_cloud_service_account_key'  # noqa: S105
GOOGLE_CLOUD_SERVICE_ACCOUNT_KEY_PATTERN = """{
  "type": "service_account",
  "project_id": "[a-z][a-z0-9-]+",
  "private_key_id": "[a-z0-9]{40}",
  "private_key": "-----BEGIN PRIVATE KEY-----\\\\n([a-zA-Z0-9+/=\\\\n]+)\\\\n-----END \
PRIVATE KEY-----\\\\n",
  "client_email": "[a-z0-9._%+-]+@[a-z0-9-]+\\.iam\\.gserviceaccount\\.com",
  "client_id": "[0-9]{12,}",
  "auth_uri": "https://accounts.google.com/o/oauth2/auth",
  "token_uri": "https://oauth2.googleapis.com/token",
  "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
  "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/[a-z0-9-]+%40[a-z0-9-]+\\.iam\\.gserviceaccount\\.com",
  "universe_domain": "googleapis.com"
}"""

OPENAI_API_KEY_SECRET_ID = 'openai_api_key'  # noqa: S105
OPENAI_API_KEY_PATTERN = '^sk-[a-zA-Z0-9-_]{32,}$'

GROK_API_KEY_SECRET_ID = 'grok_api_key'  # noqa: S105
GROK_API_KEY_PATTERN = '^xai-[a-zA-Z0-9]{80}$'

ELEVENLABS_API_KEY_SECRET_ID = 'elevenlabs_api_key'  # noqa: S105
ELEVENLABS_API_KEY_PATTERN = '^[a-f0-9]{64}$'
ELEVENLABS_VOICE_ID = 'elevenlabs_voice_id'
ELEVENLABS_VOICE_ID_PATTERN = '^[a-zA-Z0-9-_]{20,}$'

BRAVE_SEARCH_API_KEY_SECRET_ID = 'brave_search_api_key'  # noqa: S105
BRAVE_SEARCH_API_KEY_PATTERN = '^BS[a-zA-Z0-9-_]{20,}$'

VOSK_MODEL = 'vosk-model-small-en-us-0.15'
VOSK_MODEL_URL = f'https://alphacephei.com/vosk/models/{VOSK_MODEL}.zip'
VOSK_MODEL_PATH = DATA_PATH / VOSK_MODEL
VOSK_DOWNLOAD_PATH = VOSK_MODEL_PATH.with_suffix('.zip')
VOSK_DOWNLOAD_NOTIFICATION_ID = 'assistant:download-vosk'

OLLAMA_SETUP_NOTIFICATION_ID = 'assistant:ollama:setup'

OLLAMA_ONPREM_URL_SECRET_ID = 'ollama_onprem_url'  # noqa: S105
OLLAMA_ONPREM_URL_PATTERN = r'^https?://[a-zA-Z0-9.-]+(:[0-9]+)?/?$'
OLLAMA_ONPREM_SETUP_NOTIFICATION_ID = 'assistant:ollama_onprem:setup'

PIPER_MODEL = 'en/en_US/kristin/medium/en_US-kristin-medium'
PIPER_MODEL_URL = (
    f'https://huggingface.co/rhasspy/piper-voices/resolve/0c9c5d3/{PIPER_MODEL}.onnx'
)
PIPER_MODEL_HASH = '5849957f929cbf720c258f8458692d6103fff2f0e3d3b19c8259474bb06a18d4'
PIPER_MODEL_PATH = (DATA_PATH / PIPER_MODEL).with_suffix('.onnx')
PIPER_MODEL_JSON_PATH = (DATA_PATH / PIPER_MODEL).with_suffix('.onnx.json')
PIPER_DOWNLOAD_NOTIFICATION_ID = 'speech_synthesis:download-piper'

PICOVOICE_ACCESS_KEY_SECRET_ID = 'picovoice_access_key'  # noqa: S105

DEEPGRAM_API_KEY_SECRET_ID = 'deepgram_api_key'  # noqa: S105
DEEPGRAM_API_KEY_PATTERN = '^[a-f0-9]{40}$'

CEREBRAS_API_KEY_SECRET_ID = 'cerebras_api_key'  # noqa: S105
CEREBRAS_API_KEY_PATTERN = '^csk-[a-zA-Z0-9-_]{40,}$'

ASSEMBLYAI_API_KEY_SECRET_ID = 'assemblyai_api_key'  # noqa: S105
ASSEMBLYAI_API_KEY_PATTERN = '^[a-f0-9]{32}$'

RIME_API_KEY_SECRET_ID = 'rime_api_key'  # noqa: S105
RIME_API_KEY_PATTERN = '^[a-zA-Z0-9-_]{32,}$'


ASSISTANT_MCP_SERVERS_PATH = CONFIG_PATH / 'assistant_mcp_servers'



"""A simple DNS server that responds to all queries with a fixed A record."""

from __future__ import annotations

import asyncio

from quart import Quart, request

from ubo_app.constants import (
    WEB_UI_DEBUG_MODE,
    WEB_UI_LISTEN_PORT,
)
from ubo_app.logger import get_logger

logger = get_logger('system-manager')

CAPTIVE_IP = '192.168.4.1'  # The IP your machine will have on its Wi-Fi AP interface
PORT_HTTP = WEB_UI_LISTEN_PORT  # The port for the HTTP server

REDIRECT_PATHS = [
    '/generate_204',  # Android captive check
    '/hotspot-detect.html',  # Apple captive check
    '/canonical.html',  # Firefox captive check
    '/ncsi.txt',  # Windows captive check
    '/redirect',  # Microsoft variant
    '/connecttest.txt',  # Windows 11 variant
]


async def start_redirect_server() -> None:
    """Start the HTTP server that will redirect all requests to the web UI."""
    app = Quart('ubo-app-redirect-server')
    app.debug = WEB_UI_DEBUG_MODE

    @app.route('/', defaults={'path': ''})
    @app.route('/<path:path>')
    async def _(path: str) -> tuple[str, int, dict[str, str]]:
        if request.path in REDIRECT_PATHS:
            logger.info('Redirecting to the web UI...', extra={'path': path})
            return '', 302, {'Location': f'http://{CAPTIVE_IP}:{PORT_HTTP}/'}

        if request.path == '/favicon.ico':
            return '', 404, {}

        return '', 302, {'Location': f'http://{CAPTIVE_IP}:{PORT_HTTP}/'}

    await app.run_task(
        host='0.0.0.0',  # noqa: S104
        port=80,
        debug=WEB_UI_DEBUG_MODE,
    )


def main() -> None:
    """Start the HTTP server."""
    asyncio.run(start_redirect_server())



"""Thread class that tracks the UboService starting it."""

from __future__ import annotations

import threading
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ubo_app.service_thread import UboServiceThread


class UboThread(threading.Thread):
    """Thread class that tracks the `UboService` starting it."""

    ubo_service: UboServiceThread | None = None

    def start(self) -> None:
        """Start the thread."""
        from ubo_app.utils.service import get_service

        self.ubo_service = get_service()
        super().start()



# ruff: noqa: D100, D101, D103
from __future__ import annotations

import asyncio
import re
import subprocess
from typing import TypedDict

from debouncer import DebounceOptions, debounce

from ubo_app.colors import DANGER_COLOR
from ubo_app.logger import logger
from ubo_app.store.main import store
from ubo_app.store.services.notifications import (
    Chime,
    Importance,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.rpi_connect import (
    RPiConnectDoneDownloadingAction,
    RPiConnectSetPendingAction,
    RPiConnectSetStatusAction,
    RPiConnectStartDownloadingAction,
    RPiConnectStatus,
    RPiConnectUpdateServiceStateAction,
)
from ubo_app.store.services.speech_synthesis import ReadableInformation
from ubo_app.utils.apt import is_package_installed
from ubo_app.utils.async_ import create_task
from ubo_app.utils.error_handlers import report_service_error
from ubo_app.utils.monitor_unit import is_unit_active
from ubo_app.utils.server import send_command


class ConnectServiceStatus(TypedDict):
    screen_sharing_sessions: int | None
    remote_shell_sessions: int | None


@debounce(
    wait=0.5,
    options=DebounceOptions(leading=True, trailing=False, time_window=0.5),
)
async def _check_status() -> None:
    await check_is_active()
    is_installed = await is_package_installed('rpi-connect')
    is_signed_in = None
    status_data: ConnectServiceStatus | None = None
    try:
        process = await asyncio.create_subprocess_exec(
            '/usr/bin/env',
            'rpi-connect',
            'status',
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.DEVNULL,
        )
        await asyncio.wait_for(process.wait(), timeout=3)
        if process.returncode is None:
            process.kill()
        if process.stdout and process.returncode == 0:
            output = (await process.stdout.read()).decode()

            data = re.search(
                r"""Signed in: (yes|no)
(?:Screen sharing: (allowed|unavailable)(?: \((\d+) sessions? active\))?
Remote shell: (allowed|unavailable)(?: \((\d+) sessions? active\))?)?""",
                output,
            )

            if data:
                is_signed_in = data.group(1) == 'yes'
                if is_signed_in:
                    status_data = {
                        'screen_sharing_sessions': int(data.group(3) or '0')
                        if data.group(2) == 'allowed'
                        else None,
                        'remote_shell_sessions': int(data.group(5) or '0')
                        if data.group(4) == 'allowed'
                        else None,
                    }
    except (subprocess.CalledProcessError, TimeoutError):
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    title='RPi-Connect',
                    content='Failed to get status: "status" subcommand',
                    display_type=NotificationDisplayType.STICKY,
                    color=DANGER_COLOR,
                    icon='󰜺',
                    chime=Chime.FAILURE,
                ),
            ),
        )
        report_service_error()
    logger.info(
        'Checked RPi Connect Status',
        extra=status_data,
    )
    store.dispatch(
        RPiConnectSetStatusAction(
            is_installed=is_installed,
            is_signed_in=is_signed_in,
            status=None
            if status_data is None
            else RPiConnectStatus(
                screen_sharing_sessions=status_data['screen_sharing_sessions'],
                remote_shell_sessions=status_data['remote_shell_sessions'],
            ),
        ),
    )


def install_rpi_connect() -> None:
    store.dispatch(RPiConnectStartDownloadingAction())

    async def act() -> None:
        result = await send_command(
            'package',
            'install',
            'rpi-connect',
            has_output=True,
        )

        store.dispatch(RPiConnectDoneDownloadingAction())
        if result != 'installed':
            store.dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        title='RPi-Connect',
                        content='Failed to install',
                        display_type=NotificationDisplayType.STICKY,
                        color=DANGER_COLOR,
                        icon='󰜺',
                        chime=Chime.FAILURE,
                    ),
                ),
            )
        await check_status()

    create_task(act())


def uninstall_rpi_connect() -> None:
    store.dispatch(RPiConnectSetPendingAction())

    async def act() -> None:
        result = await send_command(
            'package',
            'uninstall',
            'rpi-connect',
            has_output=True,
        )

        if result != 'uninstalled':
            store.dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        title='RPi-Connect',
                        content='Failed to uninstall',
                        display_type=NotificationDisplayType.STICKY,
                        color=DANGER_COLOR,
                        icon='󰜺',
                        chime=Chime.FAILURE,
                    ),
                ),
            )
        await check_status()

    create_task(act())


def sign_out() -> None:
    async def act() -> None:
        try:
            process = await asyncio.create_subprocess_exec(
                '/usr/bin/env',
                'rpi-connect',
                'signout',
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL,
            )
            await process.wait()
            await check_status()
        except subprocess.CalledProcessError:
            store.dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        title='RPi-Connect',
                        content='Failed to logout',
                        display_type=NotificationDisplayType.STICKY,
                        color=DANGER_COLOR,
                        icon='󰜺',
                        chime=Chime.FAILURE,
                    ),
                ),
            )
            raise

    create_task(act())


@store.with_state(lambda state: state.lightdm.is_active)
def start_service(is_lightdm_active: bool) -> None:  # noqa: FBT001
    """Start the RPi Connect service."""

    async def act() -> None:
        if not is_lightdm_active:
            store.dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        title='RPi-Connect',
                        content='LightDM is not running',
                        extra_information=ReadableInformation(
                            text="""\
LightDM is not running so RPi-Connect will run without screen sharing and you can only \
use the remote shell feature. To enable screen sharing, start LightDM service in \
Settings → Desktop → LightDM menu, then come back here to stop and start RPi-Connect \
again.""",
                            piper_text="""\
LightDM is not running so RPi-Connect will run without screen sharing and you can only \
use the remote shell feature. To enable screen sharing, start LightDM service in \
LightDM item under Desktop menu under Settings menu, then come back here to stop and \
start RPi-Connect again.""",
                            picovoice_text="""\
LightDM is not running so RPi-Connect will run without screen sharing and you can only \
use the remote shell feature. To enable screen sharing, start LightDM service in \
LightDM item under Desktop menu under Settings menu, then come back here to stop and \
start RPi-Connect again.""",
                        ),
                        display_type=NotificationDisplayType.STICKY,
                        importance=Importance.LOW,
                    ),
                ),
            )
        process = await asyncio.create_subprocess_exec(
            '/usr/bin/env',
            'systemctl',
            '--user',
            'start',
            'rpi-connect',
        )
        await process.wait()
        await check_status()

    create_task(act())


def stop_service() -> None:
    """Stop the RPi Connect service."""

    async def act() -> None:
        process = await asyncio.create_subprocess_exec(
            '/usr/bin/env',
            'systemctl',
            '--user',
            'stop',
            'rpi-connect',
        )
        await process.wait()
        await check_status()

    create_task(act())


async def check_is_active() -> None:
    """Check if the SSH service is active."""
    if await is_unit_active('rpi-connect', is_user_service=True):
        store.dispatch(RPiConnectUpdateServiceStateAction(is_active=True))
    else:
        store.dispatch(RPiConnectUpdateServiceStateAction(is_active=False))


async def check_status() -> None:
    await _check_status()



"""Constants of the ubo app."""

import os
from pathlib import Path

import platformdirs
from str_to_bool import str_to_bool

if __package__ is None:
    msg = 'The package name is not set.'
    raise ValueError(msg)
PACKAGE_NAME = __package__.rpartition('.')[0]
USERNAME = os.environ.get('UBO_USERNAME', 'ubo')
INSTALLATION_PATH = os.environ.get('UBO_INSTALLATION_PATH', '/opt/ubo')

FORCE_HARDWARE = str_to_bool(os.environ.get('UBO_FORCE_HARDWARE', 'False'))

DEBUG_VISUAL = str_to_bool(os.environ.get('UBO_DEBUG_VISUAL', 'False'))
DEBUG_BETA_VERSIONS = str_to_bool(os.environ.get('UBO_DEBUG_BETA_VERSIONS', 'False'))
DEBUG_PDB_SIGNAL = str_to_bool(os.environ.get('UBO_DEBUG_PDB_SIGNAL', 'False'))
DEBUG_TASKS = str_to_bool(os.environ.get('UBO_DEBUG_TASKS', 'False'))
DEBUG_DOCKER = str_to_bool(os.environ.get('UBO_DEBUG_DOCKER', 'False'))
DEBUG_TEST_UUID = str_to_bool(os.environ.get('UBO_DEBUG_TEST_UUID', 'False'))
DEBUG_MENU = str_to_bool(os.environ.get('UBO_DEBUG_MENU', 'False'))
DEBUG_SCHEDULER = str_to_bool(os.environ.get('UBO_DEBUG_SCHEDULER', 'False'))

LOG_LEVEL = os.environ.get('UBO_LOG_LEVEL', 'INFO')
GUI_LOG_LEVEL = os.environ.get('UBO_GUI_LOG_LEVEL', 'INFO')
SERVICES_PATH = (
    os.environ.get('UBO_SERVICES_PATH', '').split(':')
    if os.environ.get('UBO_SERVICES_PATH')
    else []
)
SERVER_SOCKET_PATH = Path('/run/ubo').joinpath('system_manager.sock').as_posix()
DISABLED_SERVICES = os.environ.get('UBO_DISABLED_SERVICES', '')
DISABLED_SERVICES = DISABLED_SERVICES.split(',') if DISABLED_SERVICES else []
ENABLED_SERVICES = os.environ.get('UBO_ENABLED_SERVICES', '')
ENABLED_SERVICES = ENABLED_SERVICES.split(',') if ENABLED_SERVICES else []

DISABLE_GRPC = str_to_bool(os.environ.get('UBO_DISABLE_GRPC', 'False'))
GRPC_LISTEN_ADDRESS = os.environ.get('UBO_GRPC_LISTEN_ADDRESS', '127.0.0.1')
GRPC_LISTEN_PORT = int(os.environ.get('UBO_GRPC_LISTEN_PORT', '50051'))

GRPC_ENVOY_LISTEN_ADDRESS = os.environ.get('UBO_GRPC_ENVOY_LISTEN_ADDRESS', '0.0.0.0')  # noqa: S104
GRPC_ENVOY_LISTEN_PORT = int(os.environ.get('UBO_GRPC_ENVOY_LISTEN_PORT', '50052'))

# Most of these should be changed in ubo-app and ubo-system-manager simultaneously to
# avoid breaking the system.
# TODO(sassanh): Make above comment visible to the end user when a change # noqa: FIX002
# is detected in of these values.
WEB_UI_LISTEN_ADDRESS = os.environ.get('UBO_WEB_UI_LISTEN_ADDRESS', '0.0.0.0')  # noqa: S104
WEB_UI_LISTEN_PORT = int(os.environ.get('UBO_WEB_UI_LISTEN_PORT', '4321'))
WEB_UI_DEBUG_MODE = str_to_bool(os.environ.get('UBO_WEB_UI_DEBUG_MODE', 'False'))
WEB_UI_HOTSPOT_PASSWORD = os.environ.get('UBO_WEB_UI_HOTSPOT_PASSWORD', 'ubopod-setup')

UPDATE_ASSETS_PATH = Path(f'{INSTALLATION_PATH}/_update/')

SERVICES_LOOP_GRACE_PERIOD = float(
    os.environ.get('UBO_SERVICES_LOOP_GRACE_PERIOD', '0.1'),
)
MAIN_LOOP_GRACE_PERIOD = int(os.environ.get('UBO_MAIN_LOOP_GRACE_PERIOD', '1'))
STORE_GRACE_PERIOD = int(os.environ.get('UBO_STORE_GRACE_PERIOD', '1'))

# Enable it to replace UUIDs with numerical counters in tests and log the traceback
# each time a UUID is generated.

DOCKER_CREDENTIALS_TEMPLATE_SECRET_ID = 'docker_credentials:{}'  # noqa: S105

CONFIG_PATH = platformdirs.user_config_path(appname='ubo', ensure_exists=True)
SECRETS_PATH = CONFIG_PATH / '.secrets.env'
PERSISTENT_STORE_PATH = CONFIG_PATH / 'state.json'

CACHE_PATH = Path(
    os.environ.get(
        'UBO_CACHE_PATH',
        platformdirs.user_cache_path(appname='ubo', ensure_exists=True),
    ),
)
DATA_PATH = Path(
    os.environ.get(
        'UBO_DATA_PATH',
        platformdirs.user_data_path(appname='ubo', ensure_exists=True),
    ),
)

DISPLAY_BAUDRATE = int(os.environ.get('UBO_DISPLAY_BAUDRATE', '60_000_000'))
WIDTH = 240
HEIGHT = 240
BYTES_PER_PIXEL = 2

NOTIFICATIONS_FLASH_TIME = 4


CORE_SERVICE_IDS = [
    'assistant',
    'audio',
    'camera',
    'display',
    'docker',
    'ethernet',
    'file_system',
    'infrared',
    'ip',
    'keyboard',
    'keypad',
    'lightdm',
    'notifications',
    'rgb_ring',
    'rpi_connect',
    'sensors',
    'speech_recognition',
    'speech_synthesis',
    'ssh',
    'users',
    'vscode',
    'web_ui',
    'wifi',
]

TEST_INVESTIGATION_MODE = str_to_bool(
    os.environ.get('UBO_TEST_INVESTIGATION_MODE', 'False'),
)

SPEECH_RECOGNITION_FRAME_RATE = 16_000
SPEECH_RECOGNITION_SAMPLE_WIDTH = 2

DISPLAY_BLANK_TIMEOUT = 15.0  # seconds



"""Network utilities for the Raspberry Pi."""

import asyncio
import subprocess

from ubo_app.logger import logger
from ubo_app.utils import IS_RPI


async def has_gateway() -> bool:
    """Check if any network is connected."""
    try:
        # macOS uses 'route -n get default', Linux uses 'ip route'
        process = await asyncio.create_subprocess_exec(
            'which',
            'ip',
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        await process.wait()
        if process.returncode == 0:
            # For Linux
            process = await asyncio.create_subprocess_exec(
                'ip',
                'route',
                'show',
                'default',
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            await process.wait()
            if process.returncode == 0 and process.stdout:
                for line in (await process.stdout.read()).splitlines():
                    if (
                        line.startswith(b'default')
                        and b'scope link' not in line
                        and b'127.' not in line
                        and b'169.254.' not in line
                        and b'::1' not in line
                        and b'fe80::' not in line
                    ):
                        return True
        else:
            # For macOS
            process = await asyncio.create_subprocess_exec(
                'route',
                '-n',
                'get',
                'default',
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            await process.wait()
            if process.returncode == 0 and process.stdout:
                for line in (await process.stdout.read()).splitlines():
                    if b'gateway:' in line:
                        return True
    finally:
        pass
    return False


async def get_saved_wifi_ssids() -> list[str]:
    """Get a list of saved Wi-Fi SSIDs using nmcli.

    Returns:
        list: List of saved Wi-Fi SSIDs or an empty list if none are found.

    """
    if not IS_RPI:
        return []
    try:
        process = await asyncio.create_subprocess_exec(
            'nmcli',
            '-t',
            '-f',
            'NAME,TYPE',
            'connection',
            'show',
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        await process.wait()
        if process.returncode == 0 and process.stdout:
            return [
                line.split(b':')[0].decode()
                for line in (await process.stdout.read()).split(b'\n')
                if b'wifi' in line
            ]
    except subprocess.CalledProcessError as e:
        logger.exception('Error executing nmcli', extra={'output': e.stderr})
    except Exception as e:
        logger.exception('Unexpected error', extra={'error': e})

    return []



# pyright: reportMissingModuleSource=false
# ruff: noqa: D100
from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from ubo_app.utils.bus_provider import get_system_bus
from ubo_app.utils.dbus_interfaces import SystemdUnitInterface

if TYPE_CHECKING:
    from collections.abc import Callable


def to_dbus_string(string: str) -> str:
    """Convert a string to a D-Bus object path."""
    result = []
    for char in string:
        if char.isalnum() or char == '_':
            result.append(char)
        else:
            result.append(f'_{ord(char):02x}')
    return ''.join(result)


async def monitor_unit(unit_name: str, callback: Callable[[str], None]) -> None:
    """Monitor the active state of a systemd unit."""
    bus = get_system_bus()
    system_service = SystemdUnitInterface.new_proxy(
        bus=bus,
        service_name='org.freedesktop.systemd1',
        object_path=f'/org/freedesktop/systemd1/unit/{to_dbus_string(unit_name)}',
    )

    callback(await system_service.active_state)

    async for _ in system_service.properties_changed:
        active_state = await system_service.active_state
        callback(active_state)


async def is_unit_active(unit: str, *, is_user_service: bool = False) -> bool:
    """Check if the systemd unit is active."""
    process = await asyncio.create_subprocess_exec(
        '/usr/bin/env',
        'systemctl',
        *(['--user'] if is_user_service else []),
        'is-active',
        unit,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.DEVNULL,
    )
    stdout, _ = await process.communicate()
    return stdout.decode().strip() in ('active', 'activating', 'reloading')


async def is_unit_enabled(unit: str, *, is_user_service: bool = False) -> bool:
    """Check if the systemd unit is enabled."""
    process = await asyncio.create_subprocess_exec(
        '/usr/bin/env',
        'systemctl',
        *(['--user'] if is_user_service else []),
        'is-enabled',
        unit,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.DEVNULL,
    )
    stdout, _ = await process.communicate()
    return stdout.decode().strip() == 'enabled'



"""SSH service module."""

from __future__ import annotations

import asyncio
import datetime
from asyncio import Future
from typing import TYPE_CHECKING

from ubo_gui.menu.types import (
    HeadedMenu,
    HeadlessMenu,
    Menu,
    SubMenuItem,
)

from ubo_app.colors import DANGER_COLOR, SUCCESS_COLOR, WARNING_COLOR
from ubo_app.logger import logger
from ubo_app.store.core.types import (
    RegisterSettingAppAction,
    SettingsCategory,
)
from ubo_app.store.main import store
from ubo_app.store.services.notifications import (
    Importance,
    Notification,
    NotificationActionItem,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.speech_synthesis import ReadableInformation
from ubo_app.store.services.users import (
    UsersCreateUserAction,
    UsersCreateUserEvent,
    UsersDeleteUserAction,
    UsersDeleteUserEvent,
    UsersResetPasswordAction,
    UsersResetPasswordEvent,
    UsersSetUsersAction,
    UsersState,
    UserState,
)
from ubo_app.store.ubo_actions import UboDispatchItem
from ubo_app.utils.async_ import create_task
from ubo_app.utils.bus_provider import get_system_bus
from ubo_app.utils.dbus_interfaces import AccountsInterface, UserInterface
from ubo_app.utils.server import send_command

if TYPE_CHECKING:
    from ubo_app.utils.types import Subscriptions


async def create_account() -> None:
    """Create a system user account."""
    result = await send_command('users', 'create', has_output=True)
    if not result:
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    title='Failed to create account',
                    content='An error occurred while creating the user account.',
                    importance=Importance.MEDIUM,
                    icon='󰀅',
                    display_type=NotificationDisplayType.STICKY,
                    color=DANGER_COLOR,
                ),
            ),
        )

        return
    username, password = result.split(':')
    store.dispatch(
        NotificationsAddAction(
            notification=Notification(
                title='Account Info',
                content='[size=18dp][b]host:[/b] {{hostname}}\n'
                f'[b]user:[/b] {username}\n[b]pass:[/b] {password}[/size]',
                importance=Importance.MEDIUM,
                icon='󰀈',
                display_type=NotificationDisplayType.STICKY,
                extra_information=ReadableInformation(
                    text="""\
Note that in order to make ssh works for you, we had to make sure password \
authentication for ssh server is enabled, you may want to disable it later.""",
                    picovoice_text="""\
Note that in order to make ssh works for you, we had to make sure password \
authentication for {ssh|EH S EH S EY CH} server is enabled, you may want to disable it \
later.""",
                ),
                color=SUCCESS_COLOR,
            ),
        ),
    )


async def delete_account(event: UsersDeleteUserEvent) -> None:
    """Delete a user account."""
    loop = asyncio.get_running_loop()
    notification_future: Future[None] = loop.create_future()
    store.dispatch(
        NotificationsAddAction(
            notification=Notification(
                id='qrcode',
                icon='󰀕',
                title='Users',
                content=f'Delete user "{event.id}"?',
                display_type=NotificationDisplayType.STICKY,
                is_read=True,
                extra_information=ReadableInformation(
                    text='This will delete the system user account and its home '
                    'directory.',
                ),
                expiration_timestamp=datetime.datetime.now(tz=datetime.UTC),
                actions=[
                    NotificationActionItem(
                        action=lambda: loop.call_soon_threadsafe(
                            notification_future.set_result,
                            None,
                        )
                        and None,
                        icon='󰀍',
                        dismiss_notification=True,
                    ),
                ],
                show_dismiss_action=False,
                dismiss_on_close=True,
                on_close=lambda: loop.call_soon_threadsafe(notification_future.cancel),
            ),
        ),
    )

    await notification_future

    await send_command('users', 'delete', event.id)
    store.dispatch(
        NotificationsAddAction(
            notification=Notification(
                title='Users',
                content=f'User "{event.id}" deleted.',
                importance=Importance.MEDIUM,
                icon='󰀕',
                display_type=NotificationDisplayType.FLASH,
            ),
        ),
    )


async def reset_password(event: UsersResetPasswordEvent) -> None:
    """Reset the password of a user account."""
    result = await send_command('users', 'reset_password', event.id, has_output=True)
    if not result:
        store.dispatch(
            NotificationsAddAction(
                notification=Notification(
                    title='Failed to reset password',
                    content='An error occurred while resetting password for '
                    f'"{event.id}".',
                    importance=Importance.MEDIUM,
                    icon='󰀅',
                    display_type=NotificationDisplayType.STICKY,
                    color=DANGER_COLOR,
                ),
            ),
        )

        return
    username, password = result.split(':')
    store.dispatch(
        NotificationsAddAction(
            notification=Notification(
                title='Account Info',
                content='[size=18dp][b]host:[/b] {{hostname}}\n'
                f'[b]user:[/b] {username}\n[b]pass:[/b] {password}[/size]',
                importance=Importance.MEDIUM,
                icon='󰀈',
                display_type=NotificationDisplayType.STICKY,
                extra_information=ReadableInformation(
                    text="""\
Note that in order to make ssh works for you, we had to make sure password \
authentication for ssh server is enabled, you may want to disable it later.""",
                    picovoice_text="""\
Note that in order to make ssh works for you, we had to make sure password \
authentication for {ssh|EH S EH S EY CH} server is enabled, you may want to disable it \
later.""",
                ),
                color=SUCCESS_COLOR,
            ),
        ),
    )


@store.autorun(lambda state: state.users)
def users_menu(state: UsersState) -> Menu:
    """Get the SSH menu items."""
    if state.users is None:
        return HeadedMenu(
            title='󰡉Users',
            heading='Loading...',
            sub_heading='Please wait...',
            items=[],
        )
    return HeadlessMenu(
        title='󰡉Users',
        items=[
            UboDispatchItem(
                label='Add',
                icon='󰀔',
                store_action=UsersCreateUserAction(),
                background_color=WARNING_COLOR,
            ),
            *[
                SubMenuItem(
                    key=user.id,
                    label=user.id,
                    icon='󰀄',
                    sub_menu=HeadlessMenu(
                        title=user.id,
                        items=[
                            UboDispatchItem(
                                label='Reset Password',
                                icon='󰯄',
                                store_action=UsersResetPasswordAction(id=user.id),
                                background_color=WARNING_COLOR,
                            ),
                            UboDispatchItem(
                                label='Delete',
                                icon='󰀕',
                                store_action=UsersDeleteUserAction(id=user.id),
                                background_color=DANGER_COLOR,
                            ),
                        ],
                    ),
                )
                for user in state.users
            ],
        ],
    )


async def init_service() -> Subscriptions:
    """Initialize the Users service."""
    store.dispatch(
        RegisterSettingAppAction(
            priority=1,
            category=SettingsCategory.SYSTEM,
            menu_item=SubMenuItem(
                label='Users',
                icon='󰡉',
                sub_menu=users_menu,
            ),
        ),
    )

    bus = get_system_bus()
    accounts_service = AccountsInterface.new_proxy(
        bus=bus,
        service_name='org.freedesktop.Accounts',
        object_path='/org/freedesktop/Accounts',
    )

    async def get_users() -> list[UserState]:
        paths = await accounts_service.list_cached_users()
        return [
            UserState(
                id=(
                    user_name := await UserInterface.new_proxy(
                        bus=bus,
                        service_name='org.freedesktop.Accounts',
                        object_path=path,
                    ).user_name
                ),
                is_removable=user_name != 'ubo',
            )
            for path in paths
        ]

    store.dispatch(UsersSetUsersAction(users=await get_users()))

    async def monitor_user_added() -> None:
        async for path in accounts_service.user_added:
            logger.info('User added', extra={'path': path})
            store.dispatch(UsersSetUsersAction(users=await get_users()))

    async def monitor_user_deleted() -> None:
        async for path in accounts_service.user_deleted:
            logger.info('User deleted', extra={'path': path})
            store.dispatch(UsersSetUsersAction(users=await get_users()))

    return [
        store.subscribe_event(UsersCreateUserEvent, create_account),
        store.subscribe_event(UsersDeleteUserEvent, delete_account),
        store.subscribe_event(UsersResetPasswordEvent, reset_password),
        create_task(monitor_user_added()).cancel,
        create_task(monitor_user_deleted()).cancel,
    ]



# ruff: noqa: D100, D103
from __future__ import annotations

import asyncio
import pathlib
import subprocess
from typing import TYPE_CHECKING

from commands import check_status, restart, uninstall_service
from constants_ import CODE_BINARY_PATH, CODE_BINARY_URL, CODE_DOWNLOAD_PATH
from kivy.lang.builder import Builder
from kivy.properties import StringProperty
from login_page import LoginPage
from ubo_gui.menu.types import ActionItem, ApplicationItem, HeadedMenu

from ubo_app.colors import DANGER_COLOR
from ubo_app.store.core.types import RegisterSettingAppAction, SettingsCategory
from ubo_app.store.main import store
from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.vscode import (
    VSCodeDoneDownloadingAction,
    VSCodeRestartEvent,
    VSCodeStartDownloadingAction,
    VSCodeState,
    VSCodeStatus,
)
from ubo_app.store.ubo_actions import UboApplicationItem, register_application
from ubo_app.utils.async_ import create_task
from ubo_app.utils.gui import UboPageWidget
from ubo_app.utils.log_process import log_async_process


class _VSCodeQRCodePage(UboPageWidget):
    url = StringProperty()


register_application(application=_VSCodeQRCodePage, application_id='vscode:qrcode-page')
register_application(application=LoginPage, application_id='vscode:login-page')

CODE_TUNNEL_URL_PREFIX = 'https://vscode.dev/tunnel/'

if TYPE_CHECKING:
    from collections.abc import Callable

    from ubo_app.utils.types import Subscriptions


def download_code() -> None:
    CODE_BINARY_PATH.unlink(missing_ok=True)
    store.dispatch(VSCodeStartDownloadingAction())

    async def act() -> None:
        try:
            process = await asyncio.create_subprocess_exec(
                '/usr/bin/env',
                'curl',
                '-Lk',
                CODE_BINARY_URL,
                '--output',
                CODE_DOWNLOAD_PATH,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL,
            )
            await process.wait()
            store.dispatch(
                await log_async_process(process, message='Downloading VSCode'),
            )

            process = await asyncio.create_subprocess_exec(
                '/usr/bin/env',
                'tar',
                'zxf',
                CODE_DOWNLOAD_PATH,
                '-C',
                CODE_BINARY_PATH.parent,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL,
            )
            await process.wait()
            store.dispatch(await log_async_process(process, message='Unpacking VSCode'))

            process = await asyncio.create_subprocess_exec(
                CODE_BINARY_PATH,
                'version',
                'use',
                'stable',
                '--install-dir',
                CODE_BINARY_PATH,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL,
            )
            await process.wait()
            store.dispatch(
                await log_async_process(process, message='Installing VSCode'),
            )
        except subprocess.CalledProcessError:
            store.dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        id='vscode:download',
                        title='VSCode',
                        content='Failed to download',
                        display_type=NotificationDisplayType.STICKY,
                        color=DANGER_COLOR,
                        icon='󰜺',
                        chime=Chime.FAILURE,
                    ),
                ),
            )
            CODE_BINARY_PATH.unlink(missing_ok=True)
            raise
        finally:
            CODE_DOWNLOAD_PATH.unlink(missing_ok=True)
            store.dispatch(VSCodeDoneDownloadingAction())
            await check_status()

    create_task(act())


def logout() -> None:
    async def act() -> None:
        try:
            process = await asyncio.create_subprocess_exec(
                CODE_BINARY_PATH.as_posix(),
                'tunnel',
                '--accept-server-license-terms',
                'user',
                'logout',
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.DEVNULL,
            )
            await process.wait()
            await uninstall_service()
        except subprocess.CalledProcessError:
            store.dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        id='vscode:logout',
                        title='VSCode',
                        content='Failed to logout',
                        display_type=NotificationDisplayType.STICKY,
                        color=DANGER_COLOR,
                        icon='󰜺',
                        chime=Chime.FAILURE,
                    ),
                ),
            )
            raise

    create_task(act())


def status_based_actions(status: VSCodeStatus) -> list[ActionItem | ApplicationItem]:
    actions = []

    if status.is_running:
        actions.append(
            UboApplicationItem(
                label='Show URL',
                icon='󰐲',
                application_id='vscode:qrcode-page',
                initialization_kwargs={'url': f'{CODE_TUNNEL_URL_PREFIX}{status.name}'},
            ),
        )
    return actions


def login_actions(*, is_logged_in: bool | None) -> list[ActionItem | ApplicationItem]:
    actions = []
    if is_logged_in:
        actions.extend(
            [
                ActionItem(
                    label='Logout',
                    icon='󰍃',
                    action=logout,
                ),
            ],
        )
    elif is_logged_in is False:
        actions.append(
            UboApplicationItem(
                label='Login',
                icon='󰍂',
                application_id='vscode:login-page',
            ),
        )
    return actions


def generate_actions(state: VSCodeState) -> list[ActionItem | ApplicationItem]:
    actions = []
    if not state.is_pending and not state.is_downloading:
        if state.is_binary_installed:
            if state.is_logged_in and state.status:
                actions.extend(status_based_actions(state.status))
            actions.extend(login_actions(is_logged_in=state.is_logged_in))

        actions.append(
            ActionItem(
                label='Redownload Code'
                if state.is_binary_installed
                else 'Download Code CLI',
                icon='󰇚',
                action=download_code,
            ),
        )
    return actions


@store.autorun(lambda state: state.vscode)
def vscode_menu(state: VSCodeState) -> HeadedMenu:
    actions = generate_actions(state)

    status = ''
    if state.is_pending:
        status = '[size=48dp][/size]'
    elif state.status:
        if state.status.is_running:
            if state.status.name:
                status = f'Service is running, name:\n{state.status.name}'
            else:
                status = 'Service is running\nWaiting for name...'
        elif not state.status.is_service_installed:
            status = 'Service not installed'
        else:
            status = 'Service installed but not running'
    elif state.is_downloading:
        status = 'Downloading...'
    elif not state.is_binary_installed:
        status = 'Code CLI not installed'
    elif state.is_logged_in is None:
        status = 'Checking status...'
    elif state.is_logged_in is False:
        status = 'Needs authentication'
    else:
        status = 'Unknown status'

    return HeadedMenu(
        title='󰨞VSCode',
        heading='VSCode Remote Tunnel',
        sub_heading=status,
        items=actions,
        placeholder='',
    )


def generate_vscode_menu() -> Callable[[], HeadedMenu]:
    create_task(check_status())
    return vscode_menu


async def _monitor_status(end_event: asyncio.Event) -> None:
    while not end_event.is_set():
        await check_status()
        await asyncio.sleep(1)


async def init_service() -> Subscriptions:
    store.dispatch(
        RegisterSettingAppAction(
            menu_item=ActionItem(label='VSCode', icon='󰨞', action=generate_vscode_menu),
            category=SettingsCategory.REMOTE,
        ),
    )

    await check_status()

    end_event = asyncio.Event()
    create_task(_monitor_status(end_event))

    return [
        store.subscribe_event(VSCodeRestartEvent, restart),
        end_event.set,
    ]


Builder.load_file(
    pathlib.Path(__file__)
    .parent.joinpath('vscode_qrcode_page.kv')
    .resolve()
    .as_posix(),
)



"""Module for sending commands to the system manager socket."""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Literal, overload

from ubo_app.constants import SERVER_SOCKET_PATH
from ubo_app.logger import logger
from ubo_app.utils import IS_RPI

if TYPE_CHECKING:
    from collections.abc import AsyncIterator


@overload
async def send_command(*command: str) -> None: ...
@overload
async def send_command(
    *command: str,
    has_output: Literal[True],
) -> str: ...
@overload
async def send_command(
    *command: str,
    has_output_stream: Literal[True],
) -> AsyncIterator[str]: ...
async def send_command(
    *command_: str,
    has_output: bool = False,
    has_output_stream: bool = False,
) -> AsyncIterator[str] | str | None:
    """Send a command to the system manager socket."""
    if not IS_RPI:
        return None

    command = ' '.join(command_)

    try:
        reader, writer = await asyncio.open_unix_connection(SERVER_SOCKET_PATH)

        logger.debug('Sending command:', extra={'command': command})

        writer.write(command.encode() + b'\0')
        if has_output:
            response = ''
            datagram = (await reader.readuntil(b'\0'))[:-1]
            if datagram:
                response = datagram.decode('utf-8')
                logger.debug('Server response:', extra={'response': response})

            writer.close()
            return response

        if has_output_stream:

            async def generator() -> AsyncIterator[str]:
                while datagram := (await reader.readuntil(b'\0'))[:-1]:
                    yield datagram.decode('utf-8')
                    logger.debug(
                        'Server response:',
                        extra={
                            'command': command,
                            'response': datagram.decode('utf-8'),
                        },
                    )

                writer.close()

            return generator()

    except Exception:
        logger.exception(
            'Failed to send command to the server',
            extra={'command': command},
        )
        raise



"""Provides reusable gui stuff."""

from __future__ import annotations

import pathlib
import uuid
from enum import StrEnum
from typing import TYPE_CHECKING, Literal, TypeAlias

from kivy.clock import Clock
from kivy.graphics.texture import Texture
from kivy.lang.builder import Builder
from kivy.metrics import dp
from kivy.properties import (
    AliasProperty,
    NumericProperty,
    ObjectProperty,
    StringProperty,
)
from ubo_gui.menu.types import ActionItem, Item
from ubo_gui.page import PageWidget
from ubo_gui.prompt import PromptWidget
from ubo_gui.utils import mainthread_if_needed

from ubo_app.colors import SUCCESS_COLOR
from ubo_app.constants import HEIGHT, WIDTH

if TYPE_CHECKING:
    from kivy.uix.widget import Widget

ZOOM_FACTOR = 1.1
SCROLL_STEP = 10
ItemParameters: TypeAlias = dict[Literal['background_color', 'color', 'icon'], str]

SELECTED_ITEM_PARAMETERS: ItemParameters = {
    'background_color': SUCCESS_COLOR,
    'icon': '󰱒',
}
UNSELECTED_ITEM_PARAMETERS: ItemParameters = {
    'icon': '󰄱',
}


class UboPageWidget(PageWidget):
    """Base class for all UBO pages."""

    id: str

    def __init__(self, **kwargs: object) -> None:
        """Initialize the UBO page widget."""
        self.id = uuid.uuid4().hex
        kwargs = {**kwargs}
        items = kwargs.pop('items', None)
        if items is not None and not isinstance(items, list):
            msg = 'items must be a list'
            raise TypeError(msg)
        super().__init__(items=items, **kwargs)


class UboPromptWidget(PromptWidget, UboPageWidget):
    """Base class for all UBO prompts."""


class RawTextViewer(UboPageWidget):
    """Kivy widget for displaying text in a scrollable view."""

    text: str = StringProperty()

    def go_up(self) -> None:
        """Scroll up the error report."""
        self.ids.scrollable_widget.y = max(
            self.ids.scrollable_widget.y - dp(100),
            self.ids.container.y
            - (self.ids.scrollable_widget.height - self.ids.container.height),
        )

    def go_down(self) -> None:
        """Scroll down the error report."""
        self.ids.scrollable_widget.y = min(
            self.ids.scrollable_widget.y + dp(100),
            self.ids.container.y,
        )


class ScrollControl(StrEnum):
    """Enum for scroll control."""

    VERTICAL = 'vertical_scroll'
    HORIZONTAL = 'horizontal_scroll'
    ZOOM = 'zoom_scroll'


class RawImageViewer(UboPageWidget):
    """Kivy widget for displaying raw image."""

    def _get_texture(self) -> None:
        """Update the image when the image property changes."""
        texture = Texture.create(
            size=(self.width, self.height),
            colorfmt='rgb',
        )
        texture.blit_buffer(
            self.image,
            colorfmt='rgb',
            bufferfmt='ubyte',
        )
        texture.flip_vertical()

        return texture

    active_control: ScrollControl = StringProperty(ScrollControl.VERTICAL)
    width: int = NumericProperty()
    height: int = NumericProperty()
    image: bytes = ObjectProperty()
    texture: Texture = AliasProperty(getter=_get_texture, bind=['image'])

    def on_texture(self, instance: RawImageViewer, texture: Texture) -> None:
        """Reset position based on the size of the new texture."""
        _ = instance, texture
        Clock.schedule_once(self._center)

    def on_kv_post(self, base_widget: Widget) -> None:
        """Set position based on the size of the new texture."""
        _ = base_widget
        Clock.schedule_once(self._center)

    def on_size(self, instance: RawImageViewer, value: tuple[int, int]) -> None:
        """Center the image based on new container size."""
        _ = instance, value
        Clock.schedule_once(self._center)

    @mainthread_if_needed
    def _center(self, _: float = 0) -> None:
        zoom_factor = min(
            self.ids.container.width / self.ids.scrollable_widget.width,
            self.ids.container.height / self.ids.scrollable_widget.height,
        )
        self.ids.scrollable_widget.width *= zoom_factor
        self.ids.scrollable_widget.height *= zoom_factor
        self.ids.scrollable_widget.x = (
            self.ids.container.width - self.ids.scrollable_widget.width
        ) / 2
        self.ids.scrollable_widget.y = (
            self.ids.container.height - self.ids.scrollable_widget.height
        ) / 2

    @mainthread_if_needed
    def _apply_limits(self, _: float = 0) -> None:
        self.ids.scrollable_widget.x = min(
            max(
                self.ids.scrollable_widget.x,
                self.ids.container.x
                + self.ids.container.width / 2
                - self.ids.scrollable_widget.width,
            ),
            self.ids.container.x + self.ids.container.width / 2,
        )
        self.ids.scrollable_widget.y = min(
            max(
                self.ids.scrollable_widget.y,
                self.ids.container.y
                + self.ids.container.height / 2
                - self.ids.scrollable_widget.height,
            ),
            self.ids.container.y + self.ids.container.height / 2,
        )

    def _zoom(self, factor: float) -> None:
        center = (
            (
                self.ids.scrollable_widget.x
                - self.ids.container.x
                - self.ids.container.width / 2
            )
            / self.ids.scrollable_widget.width,
            (
                self.ids.scrollable_widget.y
                - self.ids.container.y
                - self.ids.container.height / 2
            )
            / self.ids.scrollable_widget.height,
        )
        self.ids.scrollable_widget.width = max(
            min(self.ids.scrollable_widget.width * factor, dp(self.width) * 10),
            dp(WIDTH) / 2,
        )
        self.ids.scrollable_widget.height = max(
            min(self.ids.scrollable_widget.height * factor, dp(self.height) * 10),
            dp(HEIGHT) / 2,
        )
        self.ids.scrollable_widget.pos = (
            center[0] * self.ids.scrollable_widget.width
            + self.ids.container.x
            + self.ids.container.width / 2,
            center[1] * self.ids.scrollable_widget.height
            + self.ids.container.y
            + self.ids.container.height / 2,
        )

    def go_up(self) -> None:
        """Scroll up, left or zoom in the image based on the active control."""
        match self.active_control:
            case ScrollControl.VERTICAL:
                self.ids.scrollable_widget.y -= dp(SCROLL_STEP)
            case ScrollControl.HORIZONTAL:
                self.ids.scrollable_widget.x -= dp(SCROLL_STEP)
            case ScrollControl.ZOOM:
                self._zoom(ZOOM_FACTOR)
        self._apply_limits()

    def go_down(self) -> None:
        """Scroll down, right or zoom out the image based on the active control."""
        match self.active_control:
            case ScrollControl.VERTICAL:
                self.ids.scrollable_widget.y += dp(SCROLL_STEP)
            case ScrollControl.HORIZONTAL:
                self.ids.scrollable_widget.x += dp(SCROLL_STEP)
            case ScrollControl.ZOOM:
                self._zoom(1 / ZOOM_FACTOR)
        self._apply_limits()

    def _activate_vertical_scroll(self) -> None:
        self.active_control = ScrollControl.VERTICAL

    def _activate_horizontal_scroll(self) -> None:
        self.active_control = ScrollControl.HORIZONTAL

    def _activate_zoom_scroll(self) -> None:
        self.active_control = ScrollControl.ZOOM

    def get_item(self, index: int) -> Item | None:
        """Get the item at the given index."""
        match index:
            case 0:
                return ActionItem(action=self._activate_vertical_scroll)

            case 1:
                return ActionItem(action=self._activate_horizontal_scroll)

            case 2:
                return ActionItem(action=self._activate_zoom_scroll)


Builder.load_file(
    pathlib.Path(__file__).parent.joinpath('raw_text_viewer.kv').resolve().as_posix(),
)
Builder.load_file(
    pathlib.Path(__file__).parent.joinpath('raw_image_viewer.kv').resolve().as_posix(),
)



# ruff: noqa: D100, D101, D102
from __future__ import annotations

from typing import TypeAlias

from immutable import Immutable
from redux import BaseAction, BaseEvent

RgbColorElement: TypeAlias = float | int
RgbColor: TypeAlias = (
    tuple[RgbColorElement, RgbColorElement, RgbColorElement]
    | tuple[RgbColorElement, RgbColorElement, RgbColorElement, RgbColorElement]
)


class RgbRingAction(BaseAction): ...


class RgbRingEvent(BaseEvent): ...


class RgbRingSetIsBusyAction(RgbRingAction):
    is_busy: bool | None = None


class RgbRingCommandAction(RgbRingAction):
    def as_command(self) -> str:
        return ''


class RgbRingWaitableCommandAction(RgbRingCommandAction):
    wait: int = 100

    def as_command(self) -> str:
        return f'{super().as_command()} {self.wait}'


class RgbRingColorfulCommandAction(RgbRingCommandAction):
    color: RgbColor = (255, 255, 255)

    def as_command(self) -> str:
        return f'{super().as_command()} {self.color[0]} {self.color[1]} {self.color[2]}'


class RgbRingSetEnabledAction(RgbRingCommandAction):
    enabled: bool = True

    def as_command(self) -> str:
        if self.enabled:
            return 'set_enabled'
        return 'set_disabled'


class RgbRingSetAllAction(RgbRingColorfulCommandAction):
    def as_command(self) -> str:
        return f'set_all{super().as_command()}'


class RgbRingSetBrightnessAction(RgbRingCommandAction):
    brightness: float = 0.5

    def as_command(self) -> str:
        if 0 <= self.brightness <= 1:
            return f'set_brightness {self.brightness}'
        msg = f'Invalid brightness value: {self.brightness}. Must be between 0 and 1.'
        raise ValueError(msg)


class RgbRingBlankAction(RgbRingCommandAction):
    def as_command(self) -> str:
        return 'blank'


class RgbRingRainbowAction(RgbRingWaitableCommandAction):
    wait: int = 1
    rounds: int

    def as_command(self) -> str:
        return f'rainbow {self.rounds}{super().as_command()}'


class RgbRingProgressWheelStepAction(RgbRingColorfulCommandAction):
    def as_command(self) -> str:
        return f'progress_wheel_step{super().as_command()}'


class RgbRingPulseAction(RgbRingWaitableCommandAction, RgbRingColorfulCommandAction):
    repetitions: int = 5

    def as_command(self) -> str:
        return f'pulse{super().as_command()} {self.repetitions}'


class RgbRingBlinkAction(RgbRingWaitableCommandAction, RgbRingColorfulCommandAction):
    repetitions: int = 5

    def as_command(self) -> str:
        return f'blink{super().as_command()} {self.repetitions}'


class RgbRingSpinningWheelAction(
    RgbRingWaitableCommandAction,
    RgbRingColorfulCommandAction,
):
    length: int = 5
    repetitions: int = 5

    def as_command(self) -> str:
        return f'spinning_wheel{super().as_command()} {self.length} {self.repetitions}'


class RgbRingProgressWheelAction(RgbRingColorfulCommandAction):
    percentage: float = 0.5

    def as_command(self) -> str:
        return f'progress_wheel{super().as_command()} {self.percentage}'


class RgbRingFillUptoAction(RgbRingWaitableCommandAction, RgbRingColorfulCommandAction):
    percentage: float = 0.5

    def as_command(self) -> str:
        return f'fill_upto{super().as_command()} {self.percentage}'


class RgbRingFillDownfromAction(
    RgbRingWaitableCommandAction,
    RgbRingColorfulCommandAction,
):
    percentage: float = 0.5

    def as_command(self) -> str:
        return f'fill_downfrom{super().as_command()} {self.percentage}'


class RgbRingSequenceAction(RgbRingCommandAction):
    sequence: list[RgbRingCommandAction]

    def as_command(self) -> str:
        return ' | '.join(command.as_command() for command in self.sequence)


class RgbRingCommandEvent(RgbRingEvent):
    command: list[str]


class RgbRingState(Immutable):
    is_busy: bool



# ruff: noqa: D100, D101
from __future__ import annotations

from immutable import Immutable
from redux import BaseAction, BaseEvent


class UsersAction(BaseAction): ...


class UsersSetUsersAction(UsersAction):
    users: list[UserState]


class UsersCreateUserAction(UsersAction): ...


class UsersDeleteUserAction(UsersAction):
    id: str


class UsersResetPasswordAction(UsersAction):
    id: str


class UsersEvent(BaseEvent): ...


class UsersCreateUserEvent(UsersEvent): ...


class UsersDeleteUserEvent(UsersEvent):
    id: str


class UsersResetPasswordEvent(UsersEvent):
    id: str


class UserState(Immutable):
    id: str
    is_removable: bool


class UsersState(Immutable):
    users: list[UserState] | None = None



"""Installed versions of ubo."""

from __future__ import annotations

from pathlib import Path

from ubo_app.constants import INSTALLATION_PATH


def get_installed_versions() -> list[Path]:
    """Get a list of installed versions of ubo."""
    installation_path = Path(INSTALLATION_PATH)
    if not installation_path.exists():
        return []

    return [
        item
        for item in Path(INSTALLATION_PATH).iterdir()
        if item.is_dir()
        and not item.is_symlink()
        and (item / 'bin' / 'activate').is_file()
    ]



# ruff: noqa: D100, D103
from __future__ import annotations

import json
import logging
import logging.handlers
import os
import sys
from typing import TYPE_CHECKING, ClassVar, cast

if TYPE_CHECKING:
    from collections.abc import Mapping

    from ubo_app.utils.types import Subscriptions

VERBOSE = 5


COLORS_HEX = {
    logging.NOTSET: '#545454',  # Very Dark Gray
    VERBOSE: '#848484',  # Dark Gray
    logging.DEBUG: '#b6b6b6',  # Light Gray
    logging.INFO: '#64C864',  # Green
    logging.WARNING: '#FFA500',  # Orange
    logging.ERROR: '#FF9696',  # Red
    logging.CRITICAL: '#FF00FF',  # Magenta
}
COLORS_ANSI_RGB = {
    logging.NOTSET: '\033[38;2;84;84;84m',  # Very Dark Gray (RGB)
    VERBOSE: '\033[38;2;132;132;132m',  # Dark Gray (RGB)
    logging.DEBUG: '\033[38;2;182;182;182m',  # Light Gray (RGB)
    logging.INFO: '\033[38;2;100;200;100m',  # Green (RGB)
    logging.WARNING: '\033[38;2;255;165;0m',  # Orange (RGB)
    logging.ERROR: '\033[38;2;255;150;150m',  # Red (RGB)
    logging.CRITICAL: '\033[38;2;255;0;255m',  # Magenta (RGB)
}
COLORS_ANSI_BASIC = {
    logging.NOTSET: '\033[90m',  # Very Dark Gray (ANSI)
    VERBOSE: '\033[37m',  # Dark Gray (ANSI)
    logging.DEBUG: '\033[97m',  # Light Gray (ANSI)
    logging.INFO: '\033[92m',  # Green (ANSI)
    logging.WARNING: '\033[93m',  # Yellow (ANSI)
    logging.ERROR: '\033[91m',  # Red (ANSI)
    logging.CRITICAL: '\033[95m',  # Magenta (ANSI)
}
ANSI_RESET = '\033[0m'  # Reset color


def handle_circular_references(
    obj: object,
    seen: dict[int, str | dict | list | tuple] | None = None,
) -> object:
    if isinstance(obj, (str, int, float, bool)) or obj is None:
        return obj

    if seen is None:
        seen = {}

    obj_id = id(obj)
    if obj_id in seen:
        return seen[obj_id]

    seen[obj_id] = '<circular reference>'

    if isinstance(obj, dict):
        result = {
            key: handle_circular_references(value, seen) for key, value in obj.items()
        }
    elif isinstance(obj, list):
        result = [handle_circular_references(item, seen) for item in obj]
    elif isinstance(obj, tuple):
        result = tuple(handle_circular_references(item, seen) for item in obj)
    else:
        result = str(obj)

    seen[obj_id] = result

    return result


class UboLogger(logging.getLoggerClass()):
    def __init__(self: UboLogger, name: str, level: int = logging.NOTSET) -> None:
        super().__init__(name, level)

        logging.addLevelName(VERBOSE, 'VERBOSE')

    def verbose(
        self: UboLogger,
        msg: object,
        *args: tuple[object],
        exc_info: logging._ExcInfoType | None = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Mapping[str, object] | None = None,
    ) -> None:
        if self.isEnabledFor(VERBOSE):
            self._log(
                VERBOSE,
                msg,
                args=args,
                exc_info=exc_info,
                stack_info=stack_info,
                stacklevel=stacklevel,
                extra=extra,
            )


logging.setLoggerClass(UboLogger)


def get_logger(name: str) -> UboLogger:
    logger = cast('UboLogger', logging.getLogger(name))
    logger.propagate = False
    return logger


def supports_truecolor() -> bool:
    """Check if the terminal supports truecolor (24-bit)."""
    color_term = os.environ.get('COLORTERM', '')
    return color_term == 'truecolor'


logger = get_logger('ubo-app')


class ExtraFormatter(logging.Formatter):
    def_keys = (
        'name',
        'msg',
        'args',
        'levelname',
        'levelno',
        'pathname',
        'filename',
        'module',
        'exc_info',
        'exc_text',
        'stack_info',
        'lineno',
        'funcName',
        'created',
        'msecs',
        'relativeCreated',
        'thread',
        'threadName',
        'processName',
        'process',
        'message',
    )

    def format(self: ExtraFormatter, record: logging.LogRecord) -> str:
        string = super().format(record)
        extra = {k: v for k, v in record.__dict__.items() if k not in self.def_keys}

        if len(extra) > 0:
            string += ' - ' + json.dumps(
                handle_circular_references(extra),
                sort_keys=True,
                indent=2,
                default=str,
            ).replace('\\n', '\n')

        return string


class StdOutExtraFormatter(ExtraFormatter):
    max_length: int | None = None

    def format(self: StdOutExtraFormatter, record: logging.LogRecord) -> str:
        string = super().format(record)

        if self.max_length and len(string) > self.max_length:
            string = string[: self.max_length - 3] + '...'

        # Get the color for the log level and apply it
        color = (COLORS_ANSI_RGB if supports_truecolor() else COLORS_ANSI_BASIC).get(
            record.levelno,
            ANSI_RESET,
        )
        return f'{color}{string}{ANSI_RESET}'


class ThreadLevelFilter(logging.Filter):
    thread_levels: ClassVar[dict[str, int]] = {}

    @classmethod
    def set_thread_level(cls, thread_name: str, level: int | None) -> None:
        if level is None:
            if thread_name in cls.thread_levels:
                del cls.thread_levels[thread_name]
        else:
            cls.thread_levels[thread_name] = level

    def filter(self: ThreadLevelFilter, record: logging.LogRecord) -> bool:
        thread_name = record.threadName
        if thread_name is None:
            return True
        level = self.thread_levels.get(thread_name, logging.NOTSET)
        return record.levelno >= level


thread_level_filter = ThreadLevelFilter()


def add_stdout_handler(
    logger: UboLogger,
    level: int = logging.DEBUG,
) -> Subscriptions:
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(level)
    formatter = StdOutExtraFormatter(
        '%(created)f [%(levelname)s] %(message)s',
        '%Y-%m-%d %H:%M:%S',
    )
    if level <= logging.DEBUG:
        formatter.max_length = 10000
    stdout_handler.setFormatter(formatter)
    stdout_handler.addFilter(thread_level_filter)
    logger.addHandler(stdout_handler)

    def cleanup() -> None:
        stdout_handler.flush()
        stdout_handler.close()
        logger.removeHandler(stdout_handler)

    return [cleanup]


def add_file_handler(
    logger: UboLogger,
    level: int = logging.DEBUG,
) -> Subscriptions:
    file_handler = logging.handlers.RotatingFileHandler(
        f'{logger.name}.log',
        backupCount=3,
        maxBytes=1_000_000,
    )
    file_handler.setLevel(level)
    file_handler.setFormatter(
        ExtraFormatter(
            '%(created)f [%(levelname)s] %(message)s',
            '%Y-%m-%d %H:%M:%S',
        ),
    )
    file_handler.addFilter(thread_level_filter)
    logger.addHandler(file_handler)

    def cleanup() -> None:
        file_handler.flush()
        file_handler.close()
        logger.removeHandler(file_handler)

    return [cleanup]


def get_log_level() -> int | None:
    from ubo_app.constants import LOG_LEVEL

    if LOG_LEVEL:
        import logging

        return logging.getLevelNamesMapping()[LOG_LEVEL]
    return None


def get_gui_log_level() -> int | None:
    from ubo_app.constants import GUI_LOG_LEVEL

    if GUI_LOG_LEVEL:
        import logging

        return logging.getLevelNamesMapping()[GUI_LOG_LEVEL]
    return None


def setup_loggers() -> Subscriptions:
    level = get_log_level()
    subscriptions: Subscriptions = []

    if level is not None:
        logger.setLevel(level)
        subscriptions.extend(add_file_handler(logger, level))
        subscriptions.extend(add_stdout_handler(logger, level))

    gui_level = get_gui_log_level()

    if gui_level is not None:
        import ubo_gui.logger

        ubo_gui.logger.logger.setLevel(gui_level)
        ubo_gui.logger.add_file_handler(gui_level)
        ubo_gui.logger.add_stdout_handler(gui_level)

    return subscriptions


__all__ = ('add_file_handler', 'add_stdout_handler', 'logger', 'setup_loggers')



"""Sensors reducer."""

from __future__ import annotations

from dataclasses import replace

from redux import BaseEvent, InitAction, InitializationActionError, ReducerResult

from ubo_app.store.services.sensors import (
    Sensor,
    SensorsAction,
    SensorsReportReadingAction,
    SensorsState,
    SensorState,
)

Action = InitAction | SensorsAction


def reducer(
    state: SensorsState | None,
    action: Action,
) -> ReducerResult[SensorsState, Action, BaseEvent]:
    """Sensors reducer."""
    if state is None:
        if isinstance(action, InitAction):
            return SensorsState()
        raise InitializationActionError(action)

    match action:
        case SensorsReportReadingAction(sensor=Sensor.TEMPERATURE):
            return replace(state, temperature=SensorState(value=action.reading))
        case SensorsReportReadingAction(sensor=Sensor.LIGHT):
            return replace(state, light=SensorState(value=action.reading))
        case _:
            return state



"""Menu items for the system menu."""

from __future__ import annotations

from ubo_gui.menu.types import HeadlessMenu, SubMenuItem

from ubo_app.store.main import store
from ubo_app.store.services.audio import AudioInstallDriverAction
from ubo_app.store.services.camera import (
    CameraInstallDriverAction,
    CameraRestoreDefaultAction,
)
from ubo_app.store.settings.services import service_items
from ubo_app.store.settings.types import (
    SettingsToggleBetaVersionsAction,
    SettingsTogglePdbSignalAction,
    SettingsToggleVisualDebugAction,
)
from ubo_app.store.ubo_actions import UboDispatchItem
from ubo_app.utils import IS_RPI
from ubo_app.utils.eeprom import get_eeprom_data


@store.autorun(lambda state: state.settings.pdb_signal)
def _pdb_debug_icon(pdb_signal: bool) -> str:  # noqa: FBT001
    return '󰱒' if pdb_signal else '󰄱'


@store.autorun(lambda state: state.settings.visual_debug)
def _visual_debug_icon(visual_debug: bool) -> str:  # noqa: FBT001
    return '󰱒' if visual_debug else '󰄱'


@store.autorun(lambda state: state.settings.beta_versions)
def _beta_versions_icon(beta_versions: bool) -> str:  # noqa: FBT001
    return '󰱒' if beta_versions else '󰄱'


THIRD_PARTY_ITEMS = []

eeprom_data = get_eeprom_data()
if (
    (speakers := eeprom_data.get('speakers')) is not None
    and speakers.get('model') == 'wm8960'
):
    THIRD_PARTY_ITEMS.append(
        UboDispatchItem(
            label='Re/Install Audio',
            icon='',
            store_action=AudioInstallDriverAction(),
        ),
    )

if IS_RPI:
    THIRD_PARTY_ITEMS.append(
        UboDispatchItem(
            label='Arducam IMX519 AF',
            icon='󰽎',
            store_action=CameraInstallDriverAction(
                make='arducam',
                model='imx519',
                variant='autofocus',
            ),
        ),
    )
    THIRD_PARTY_ITEMS.append(
        UboDispatchItem(
            label='Arducam IMX519 FF',
            icon='󰋱',
            store_action=CameraInstallDriverAction(
                make='arducam',
                model='imx519',
                variant='fixed-focus',
            ),
        ),
    )
    THIRD_PARTY_ITEMS.append(
        UboDispatchItem(
            label='Default Camera',
            icon='󰄀',
            store_action=CameraRestoreDefaultAction(),
        ),
    )

SYSTEM_MENU: list[SubMenuItem] = [
    SubMenuItem(
        key='general',
        label='General',
        icon='󰒓',
        sub_menu=HeadlessMenu(
            title='󰒓General',
            items=[
                UboDispatchItem(
                    label='PDB Signal',
                    store_action=SettingsTogglePdbSignalAction(),
                    icon=_pdb_debug_icon,
                ),
                UboDispatchItem(
                    label='Visual Debug',
                    store_action=SettingsToggleVisualDebugAction(),
                    icon=_visual_debug_icon,
                ),
                UboDispatchItem(
                    label='Beta Versions',
                    store_action=SettingsToggleBetaVersionsAction(),
                    icon=_beta_versions_icon,
                ),
            ],
        ),
    ),
    SubMenuItem(
        key='services',
        label='Services',
        icon='',
        sub_menu=HeadlessMenu(
            title='Services',
            items=service_items,
        ),
    ),
    SubMenuItem(
        key='third_party',
        label='Third Party',
        icon='',
        sub_menu=HeadlessMenu(
            title='Third Party Tools',
            items=THIRD_PARTY_ITEMS,
        ),
    ),
]



"""Set up the hotspot configuration files."""

import pathlib
import subprocess
import sys

from ubo_app.constants import WEB_UI_HOTSPOT_PASSWORD
from ubo_app.utils.pod_id import get_pod_id
from ubo_app.utils.template_files import copy_templates, restore_backups


def main() -> None:
    """Set up the hotspot configuration files."""
    templates_path = pathlib.Path(__file__).parent / 'hotspot_templates'
    if sys.argv[1] == 'configure':
        subprocess.run(
            ['/usr/bin/env', 'iw', 'wlan0', 'set', 'power_save', 'off'],
            check=True,
        )
        subprocess.run(['/usr/bin/env', 'rfkill', 'unblock', 'wifi'], check=True)

        copy_templates(
            templates_path,
            variables={
                'SSID': get_pod_id(with_default=True),
                'PASSWORD': WEB_UI_HOTSPOT_PASSWORD,
            },
        )

        subprocess.run(['/bin/systemctl', 'enable', 'dhcpcd.service'], check=True)
        subprocess.run(['/bin/systemctl', 'restart', 'dhcpcd.service'], check=True)
        subprocess.run(['/bin/systemctl', 'enable', 'dnsmasq.service'], check=True)
        subprocess.run(['/bin/systemctl', 'restart', 'dnsmasq.service'], check=True)
        subprocess.run(['/bin/systemctl', 'unmask', 'hostapd.service'], check=True)
        subprocess.run(['/bin/systemctl', 'enable', 'hostapd.service'], check=True)
        subprocess.run(['/bin/systemctl', 'start', 'hostapd.service'], check=True)
    elif sys.argv[1] == 'restore':
        restore_backups(templates_path)

        with pathlib.Path('/etc/dhcpcd.conf').open('w') as file:
            file.write(
                '# Default dhcpcd configuration\n'
                '# Leave this blank for automatic configuration\n',
            )

        subprocess.run(['/bin/systemctl', 'stop', 'dhcpcd.service'], check=True)
        subprocess.run(['/bin/systemctl', 'disable', 'dhcpcd.service'], check=True)
        subprocess.run(['/bin/systemctl', 'stop', 'hostapd.service'], check=True)
        subprocess.run(['/bin/systemctl', 'disable', 'hostapd.service'], check=True)
        subprocess.run(['/bin/systemctl', 'mask', 'hostapd.service'], check=True)
        subprocess.run(['/bin/systemctl', 'stop', 'dnsmasq.service'], check=True)
        subprocess.run(['/bin/systemctl', 'disable', 'dnsmasq.service'], check=True)
        subprocess.run(['/usr/bin/env', 'nmcli', 'radio', 'wifi', 'on'], check=True)
    else:
        msg = 'Invalid argument'
        raise ValueError(msg)



"""Implement `init_service` for speech synthesis service."""

from __future__ import annotations

import struct
from asyncio import CancelledError
from typing import TYPE_CHECKING

import fasteners
import pvorca
from piper.voice import AudioChunk, PiperVoice
from redux import AutorunOptions
from ubo_gui.menu.types import ActionItem, HeadedMenu, HeadlessMenu, SubMenuItem

from ubo_app.constants.assistant import (
    PICOVOICE_ACCESS_KEY_SECRET_ID,
    PIPER_MODEL_PATH,
)
from ubo_app.engines.piper import PiperEngine
from ubo_app.store.core.types import RegisterSettingAppAction, SettingsCategory
from ubo_app.store.input.types import (
    InputFieldDescription,
    InputFieldType,
    QRCodeInputDescription,
    WebUIInputDescription,
)
from ubo_app.store.main import store
from ubo_app.store.services.audio import (
    AudioPlayAudioSampleAction,
    AudioPlayAudioSequenceAction,
    AudioSample,
)
from ubo_app.store.services.speech_synthesis import (
    ReadableInformation,
    SpeechSynthesisEngineName,
    SpeechSynthesisReadTextAction,
    SpeechSynthesisSetSelectedEngineAction,
    SpeechSynthesisSynthesizeTextEvent,
    SpeechSynthesisUpdateAccessKeyStatus,
)
from ubo_app.utils import secrets
from ubo_app.utils.async_ import create_task, to_thread
from ubo_app.utils.gui import SELECTED_ITEM_PARAMETERS, UNSELECTED_ITEM_PARAMETERS
from ubo_app.utils.input import ubo_input
from ubo_app.utils.persistent_store import register_persistent_store

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from ubo_app.utils.types import Subscriptions


_piper_engine = PiperEngine()


class _Context:
    picovoice_instance: pvorca.Orca | None = None
    piper_voice: PiperVoice | None = None
    picovoice_lock = fasteners.ReaderWriterLock()

    def cleanup(self: _Context) -> None:
        store.dispatch(SpeechSynthesisUpdateAccessKeyStatus(is_access_key_set=False))
        with self.picovoice_lock.write_lock():
            if self.picovoice_instance:
                self.picovoice_instance.delete()
                self.picovoice_instance = None

    def set_access_key(self: _Context, access_key: str) -> None:
        store.dispatch(SpeechSynthesisUpdateAccessKeyStatus(is_access_key_set=True))
        with self.picovoice_lock.write_lock():
            if access_key:
                if self.picovoice_instance:
                    self.picovoice_instance.delete()
                self.picovoice_instance = pvorca.create(access_key)

    def load_piper(self: _Context) -> None:
        if _piper_engine.is_setup:
            self.piper_voice = PiperVoice.load(PIPER_MODEL_PATH)


_context = _Context()


def input_access_key() -> None:
    """Input the Picovoice access key."""

    async def act() -> None:
        try:
            input_result = (
                await ubo_input(
                    prompt='Enter Picovoice Access Key',
                    title='Picovoice Access Key',
                    descriptions=[
                        QRCodeInputDescription(
                            pattern=r'^(?P<access_key>.*)$',
                            instructions=ReadableInformation(
                                text='Convert the Picovoice access key to a QR code '
                                'and hold it in front of the camera to scan it.',
                                picovoice_text='Convert the Picovoice access key to a '
                                '{QR|K Y UW AA R} and hold it in front of the camera '
                                'to scan it.',
                            ),
                        ),
                        WebUIInputDescription(
                            fields=[
                                InputFieldDescription(
                                    name='access_key',
                                    label='Access Key',
                                    description='Enter Picovoice Access Key',
                                    type=InputFieldType.TEXT,
                                    required=True,
                                    title='Picovoice Access Key',
                                ),
                            ],
                        ),
                    ],
                )
            )[1]
            access_key = input_result.data.get('access_key')
            if not access_key:
                return
            secrets.write_secret(key=PICOVOICE_ACCESS_KEY_SECRET_ID, value=access_key)
            to_thread(_context.set_access_key, None, access_key)
        except CancelledError:
            pass

    create_task(act())


def clear_access_key() -> None:
    """Clear the Picovoice access key."""
    secrets.clear_secret(PICOVOICE_ACCESS_KEY_SECRET_ID)
    to_thread(_context.cleanup)


@store.with_state(lambda state: state.speech_synthesis.selected_engine)
def _engine(engine: SpeechSynthesisEngineName) -> SpeechSynthesisEngineName:
    return engine


piper_cache: dict[str, list[bytes]] = {}


def _get_audio_bytes(audio_item: AudioChunk | bytes) -> bytes:
    """Extract bytes from either AudioChunk or cached bytes."""
    if isinstance(audio_item, AudioChunk):
        return audio_item.audio_int16_bytes
    return audio_item


def synthesize_and_play(event: SpeechSynthesisSynthesizeTextEvent) -> None:
    """Synthesize the text."""
    engine = _engine()
    if engine == SpeechSynthesisEngineName.PIPER:
        text = event.information.piper_text
        if not _context.piper_voice:
            return
        id = hex(hash(text))

        if text in piper_cache:
            source = piper_cache[text]
            is_first_time = False
        else:
            source = _context.piper_voice.synthesize(text=text)
            piper_cache[text] = []
            is_first_time = True

        index = 0
        for audio_chunk in source:
            if audio_chunk:
                sample = _get_audio_bytes(audio_chunk)
                if is_first_time:
                    piper_cache[text].append(sample)
                store.dispatch(
                    AudioPlayAudioSequenceAction(
                        sample=AudioSample(
                            data=sample,
                            channels=1,
                            rate=_context.piper_voice.config.sample_rate,
                            width=2,
                        ),
                        id=id,
                        index=index,
                    ),
                )
                index += 1
        store.dispatch(
            AudioPlayAudioSequenceAction(
                sample=None,
                id=id,
                index=index,
            ),
        )

    elif engine == SpeechSynthesisEngineName.PICOVOICE:
        with _context.picovoice_lock.read_lock():
            if not _context.picovoice_instance:
                return
            rate = _context.picovoice_instance.sample_rate

            audio_sequence = _context.picovoice_instance.synthesize(
                text=event.information.picovoice_text,
                speech_rate=event.speech_rate,
            )
        sample = b''.join(struct.pack('h', sample) for sample in audio_sequence[0])
        store.dispatch(
            AudioPlayAudioSampleAction(
                sample=AudioSample(
                    data=sample,
                    channels=1,
                    rate=rate,
                    width=2,
                ),
            ),
        )


@store.autorun(lambda state: state.speech_synthesis.is_access_key_set)
def _menu_items(is_access_key_set: bool | None) -> Sequence[ActionItem]:  # noqa: FBT001
    if is_access_key_set:
        return [
            ActionItem(
                label='Clear Access Key',
                icon='󰌊',
                action=clear_access_key,
            ),
        ]
    return [
        ActionItem(
            label='Set Access Key',
            icon='󰐲',
            action=input_access_key,
        ),
    ]


@store.autorun(lambda state: state.speech_synthesis.is_access_key_set)
def _menu_sub_heading(_: bool | None) -> str:  # noqa: FBT001
    return f"""Set the access key
Current value: {secrets.read_covered_secret(PICOVOICE_ACCESS_KEY_SECRET_ID)}"""


ENGINE_LABELS = {
    SpeechSynthesisEngineName.PIPER: 'Piper',
    SpeechSynthesisEngineName.PICOVOICE: 'Picovoice',
}


def create_engine_selector(engine: SpeechSynthesisEngineName) -> Callable[[], None]:
    """Select the speech synthesis engine."""

    def _engine_selector() -> None:
        store.dispatch(
            SpeechSynthesisSetSelectedEngineAction(engine_name=engine),
            SpeechSynthesisReadTextAction(
                information=ReadableInformation(
                    text={
                        SpeechSynthesisEngineName.PIPER: 'Piper speech synthesis '
                        'engine selected',
                        SpeechSynthesisEngineName.PICOVOICE: 'Picovoice speech '
                        'synthesis engine selected',
                    }[engine],
                ),
                engine=engine,
            ),
        )

    return _engine_selector


@store.autorun(
    lambda state: (
        state.speech_synthesis.selected_engine,
        state.assistant.provider_setup_status,
    ),
    options=AutorunOptions(memoization=False),
)
def _speech_synthesis_menu(
    data: tuple[SpeechSynthesisEngineName, dict[str, bool]],
) -> HeadlessMenu:
    selected_engine, _ = data

    if _piper_engine.is_setup and _context.piper_voice is None:
        to_thread(_context.load_piper)

    return HeadlessMenu(
        title='󰔊Speech Synthesis',
        items=[
            *(
                [
                    ActionItem(
                        key='download',
                        label='Setup Piper',
                        icon='󰇚',
                        action=_piper_engine.setup,
                    ),
                ]
                if not _piper_engine.is_setup
                else []
            ),
            SubMenuItem(
                key='select_engine',
                label='Select Engine',
                icon='󰔊',
                sub_menu=HeadlessMenu(
                    title=f'󰔊Select Engine: {selected_engine}',
                    items=[
                        (
                            selection_parameters := SELECTED_ITEM_PARAMETERS
                            if engine == selected_engine
                            else UNSELECTED_ITEM_PARAMETERS,
                        )
                        and ActionItem(
                            label=ENGINE_LABELS[engine],
                            action=create_engine_selector(engine),
                            key=engine.name,
                            **selection_parameters,
                        )
                        for engine in SpeechSynthesisEngineName
                        if _piper_engine.is_setup
                        or engine != SpeechSynthesisEngineName.PIPER
                    ],
                ),
            ),
        ],
    )


def init_service() -> Subscriptions:
    """Initialize speech synthesis service."""
    access_key = secrets.read_secret(PICOVOICE_ACCESS_KEY_SECRET_ID)
    if access_key:
        to_thread(_context.set_access_key, None, access_key)
    else:
        to_thread(_context.cleanup)

    register_persistent_store(
        'speech_synthesis:selected_engine',
        lambda state: state.speech_synthesis.selected_engine,
    )

    to_thread(_context.load_piper)

    store.dispatch(
        RegisterSettingAppAction(
            category=SettingsCategory.ACCESSIBILITY,
            priority=10,
            menu_item=SubMenuItem(
                label='Speech Synthesis',
                icon='󰔊',
                sub_menu=_speech_synthesis_menu,
            ),
            key='engines',
        ),
    )

    store.dispatch(
        RegisterSettingAppAction(
            category=SettingsCategory.ACCESSIBILITY,
            priority=0,
            menu_item=SubMenuItem(
                label='Picovoice Settings',
                icon='PV',
                sub_menu=HeadedMenu(
                    title='Picovoice Settings',
                    heading='Picovoice',
                    sub_heading=_menu_sub_heading,
                    items=_menu_items,
                ),
            ),
            key='settings',
        ),
    )

    return [
        store.subscribe_event(
            SpeechSynthesisSynthesizeTextEvent,
            lambda event: to_thread(synthesize_and_play, None, event),
        ),
        _context.cleanup,
    ]



# ruff: noqa: D100, D103
from __future__ import annotations

from dataclasses import replace

from kivy.utils import get_color_from_hex
from redux import (
    BaseEvent,
    CompleteReducerResult,
    FinishAction,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.services.audio import AudioPlayChimeAction
from ubo_app.store.services.notifications import (
    Importance,
    NotificationsAction,
    NotificationsAddAction,
    NotificationsClearAction,
    NotificationsClearAllAction,
    NotificationsClearByIdAction,
    NotificationsClearEvent,
    NotificationsDisplayAction,
    NotificationsDisplayEvent,
    NotificationsState,
)
from ubo_app.store.services.rgb_ring import RgbRingBlinkAction

Action = InitAction | NotificationsAction
ResultAction = RgbRingBlinkAction | AudioPlayChimeAction


def reducer(
    state: NotificationsState | None,
    action: Action,
) -> ReducerResult[NotificationsState, ResultAction, BaseEvent]:
    if state is None:
        if isinstance(action, InitAction):
            return NotificationsState(
                notifications=[],
                unread_count=0,
            )
        raise InitializationActionError(action)

    match action:
        case NotificationsAddAction():
            events = []
            events.append(NotificationsDisplayEvent(notification=action.notification))
            if action.notification in state.notifications:
                return CompleteReducerResult(state=state, events=events)
            kivy_color = get_color_from_hex(action.notification.color)
            new_notifications = (
                [
                    action.notification
                    if notification.id == action.notification.id
                    else notification
                    for notification in state.notifications
                ]
                if any(
                    notification.id == action.notification.id
                    for notification in state.notifications
                )
                else [action.notification, *state.notifications]
            )
            return CompleteReducerResult(
                state=replace(
                    state,
                    notifications=new_notifications,
                    unread_count=sum(
                        1
                        for notification in new_notifications
                        if not notification.is_read
                    ),
                    progress=sum(
                        notification.progress * notification.progress_weight
                        for notification in new_notifications
                        if notification.progress is not None
                    )
                    if any(
                        notification.progress is not None
                        for notification in new_notifications
                    )
                    else None,
                ),
                actions=[
                    *(
                        [
                            RgbRingBlinkAction(
                                color=(
                                    round(kivy_color[0] * 255),
                                    round(kivy_color[1] * 255),
                                    round(kivy_color[2] * 255),
                                ),
                                repetitions={
                                    Importance.LOW: 1,
                                    Importance.MEDIUM: 2,
                                    Importance.HIGH: 3,
                                    Importance.CRITICAL: 4,
                                }[action.notification.importance],
                                wait=400,
                            ),
                        ]
                        if action.notification.blink
                        else []
                    ),
                    *(
                        [AudioPlayChimeAction(name=action.notification.chime)]
                        if action.notification.chime
                        else []
                    ),
                ],
                events=events,
            )

        case NotificationsDisplayAction():
            return CompleteReducerResult(
                state=state,
                events=[
                    NotificationsDisplayEvent(
                        notification=action.notification,
                        index=action.index,
                        count=action.count,
                    ),
                ],
            )

        case NotificationsClearAction():
            new_notifications = [
                notification
                for notification in state.notifications
                if notification is not action.notification
            ]
            return CompleteReducerResult(
                state=replace(
                    state,
                    notifications=new_notifications,
                    unread_count=sum(
                        1
                        for notification in new_notifications
                        if not notification.is_read
                    ),
                ),
                events=[NotificationsClearEvent(notification=action.notification)],
            )

        case NotificationsClearByIdAction():
            to_be_removed = [
                notification
                for notification in state.notifications
                if notification.id == action.id
            ]
            new_notifications = [
                notification
                for notification in state.notifications
                if notification.id != action.id
            ]
            return CompleteReducerResult(
                state=replace(
                    state,
                    notifications=new_notifications,
                    unread_count=sum(
                        1
                        for notification in new_notifications
                        if not notification.is_read
                    ),
                ),
                events=[
                    NotificationsClearEvent(notification=notification)
                    for notification in to_be_removed
                ],
            )

        case NotificationsClearAllAction() | FinishAction():
            return CompleteReducerResult(
                state=replace(state, notifications=[], unread_count=0),
                events=[
                    NotificationsClearEvent(notification=notification)
                    for notification in state.notifications
                ],
            )

        case _:
            return state



# ruff: noqa: D100
from __future__ import annotations

from dataclasses import field

from immutable import Immutable
from redux import BaseAction, BaseEvent

from ubo_app.utils.persistent_store import read_from_persistent_store


class InfraredAction(BaseAction):
    """Base action for infrared service."""


class InfraredEvent(BaseEvent):
    """Base event for infrared service."""


class InfraredHandleReceivedCodeAction(InfraredAction):
    """Carries the received infrared code."""

    protocol: str
    scancode: str


class InfraredSendCodeAction(InfraredAction):
    """Action to send an infrared code."""

    protocol: str
    scancode: str


class InfraredSetShouldPropagateAction(InfraredAction):
    """Action to set the should propagate keypad actions flag."""

    should_propagate: bool


class InfraredSetShouldReceiveAction(InfraredAction):
    """Action to set the should receive keypad actions flag."""

    should_receive: bool


class InfraredSendCodeEvent(InfraredEvent):
    """Event to send an infrared code."""

    protocol: str
    scancode: str


class InfraredState(Immutable):
    """State of the infrared service."""

    should_propagate_keypad_actions: bool = field(
        default=read_from_persistent_store(
            'infrared_state:should_propagate_keypad_actions',
            default=False,
        ),
    )
    should_receive_keypad_actions: bool = field(
        default=read_from_persistent_store(
            'infrared_state:should_receive_keypad_actions',
            default=False,
        ),
    )



"""DBus interfaces."""

from __future__ import annotations

from sdbus import (  # pyright: ignore [reportMissingModuleSource]
    DbusInterfaceCommonAsync,
    dbus_method_async,
    dbus_property_async,
    dbus_signal_async,
)


class AccountsInterface(
    DbusInterfaceCommonAsync,
    interface_name='org.freedesktop.Accounts',
):
    """DBus interface for managing user accounts."""

    @dbus_method_async(result_signature='ao')
    async def list_cached_users(self: AccountsInterface) -> list[str]:
        """List the cached users."""
        raise NotImplementedError

    @dbus_signal_async('o')
    def user_added(self) -> str:
        """Signal emitted when a user is added."""
        raise NotImplementedError

    @dbus_signal_async('o')
    def user_deleted(self) -> str:
        """Signal emitted when a user is deleted."""
        raise NotImplementedError


class UserInterface(
    DbusInterfaceCommonAsync,
    interface_name='org.freedesktop.Accounts.User',
):
    """DBus interface for querying user information."""

    @dbus_property_async(property_signature='s')
    def user_name(self: UserInterface) -> str:
        """Return the user name."""
        raise NotImplementedError


class SystemdUnitInterface(
    DbusInterfaceCommonAsync,
    interface_name='org.freedesktop.systemd1.Unit',
):
    """DBus interface for managing systemd units."""

    @dbus_property_async(property_signature='s')
    def active_state(self: SystemdUnitInterface) -> str:
        """Return the active state of the unit."""
        raise NotImplementedError



# pyright: reportMissingModuleSource=false
# ruff: noqa: D100, D103

from ubo_app.utils import IS_RPI


def initialize_board() -> None:
    # Display will be initialized in the display module after importing it
    import ubo_app.display as _  # noqa: F401


def deinitalize_board() -> None:
    from ubo_app.display import display

    display.turn_off()

    if not IS_RPI:
        return

    from gpiozero.devices import _shutdown

    _shutdown()
